from __future__ import annotations

import json
import os
import shutil
import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Dict, List

import requests
from rich.console import Console
from tqdm.auto import tqdm
from huggingface_hub import snapshot_download
from huggingface_hub.utils import (
    LocalEntryNotFoundError,
    RepositoryNotFoundError,
    HfHubHTTPError,
)

from .config import WhisperConfig
from .subtitles import SubtitleDocument

DEFAULT_OPENAI_TRANSCRIBE_URL = "https://api.openai.com/v1/audio/transcriptions"


class TranscriptionError(RuntimeError):
    """Raised when audio transcription fails."""


class DownloadProgressBar(tqdm):
    """Custom tqdm progress bar with richer formatting for downloads."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        kwargs.setdefault(
            "bar_format",
            "{l_bar}{bar}| {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}]",
        )
        kwargs.setdefault("dynamic_ncols", True)
        super().__init__(*args, **kwargs)


DOWNLOAD_CONSOLE = Console()


@contextmanager
def _suppress_tqdm() -> Any:
    previous = os.environ.get("TQDM_DISABLE")
    os.environ["TQDM_DISABLE"] = "1"
    try:
        yield
    finally:
        if previous is None:
            os.environ.pop("TQDM_DISABLE", None)
        else:
            os.environ["TQDM_DISABLE"] = previous


def check_dependencies(config: WhisperConfig) -> None:
    """Check for required dependencies and determine execution mode."""
    backend = config.backend
    if backend == "api":
        return  # No local dependencies to check

    # Try to import the package first
    try:
        if backend == "local":
            import whisper  # type: ignore
        elif backend == "mlx":
            import mlx_whisper  # type: ignore
        config.execution_mode = "internal"
        return
    except ImportError:
        pass  # Package not found, try to find a CLI tool

    # If package import fails, look for a CLI tool
    cli_map = {
        "local": "whisper",
        "mlx": "mlx-whisper",
        "cpp": config.cpp_binary or "whisper-cpp",
    }
    cli_name = cli_map.get(backend)
    if not cli_name:
        return # Should not happen with validated config

    exec_path = shutil.which(cli_name)
    if exec_path:
        config.execution_mode = "external"
        config.cli_path = exec_path
        if backend == "cpp":
            config.cpp_binary = exec_path
        return

    # If neither package nor CLI is found, raise an error
    if backend == "local":
        raise TranscriptionError(
            "For the 'local' backend, you must either install 'openai-whisper' in the same "
            "environment as transub, or have the 'whisper' command-line tool in your PATH."
        )
    if backend == "mlx":
        raise TranscriptionError(
            "For the 'mlx' backend, you must either install 'mlx-whisper' in the same "
            "environment as transub, or have the 'mlx-whisper' command-line tool in your PATH."
        )
    if backend == "cpp":
        raise TranscriptionError(
            f"whisper.cpp executable '{cli_name}' not found in PATH. "
            "Please install whisper.cpp and ensure it is in your PATH."
        )


def transcribe_audio(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    """Transcribe audio file using the configured backend."""

    if config.backend == "local":
        return _transcribe_local(audio_path, config)
    if config.backend == "api":
        return _transcribe_api(audio_path, config)
    if config.backend == "cpp":
        return _transcribe_cpp(audio_path, config)
    if config.backend == "mlx":
        return _transcribe_mlx(audio_path, config)
    raise TranscriptionError(f"Unsupported whisper backend: {config.backend}")


def _ensure_local_mlx_repo(repo_id: str) -> str:
    """Ensure MLX whisper model repo is present locally and return its path."""

    # If repo_id already points to an existing directory, use it directly.
    repo_path = Path(repo_id).expanduser()
    if repo_path.exists():
        return str(repo_path)

    token = (
        os.getenv("HUGGINGFACEHUB_API_TOKEN")
        or os.getenv("HF_TOKEN")
        or os.getenv("HUGGINGFACE_TOKEN")
    )

    try:
        return snapshot_download(
            repo_id=repo_id,
            local_files_only=True,
            token=token,
        )
    except LocalEntryNotFoundError:
        DOWNLOAD_CONSOLE.print(
            f"Downloading MLX Whisper model [bold]{repo_id}[/]…"
        )
        try:
            return snapshot_download(
                repo_id=repo_id,
                resume_download=True,
                tqdm_class=DownloadProgressBar,
                token=token,
            )
        except RepositoryNotFoundError as err:
            raise TranscriptionError(
                "Unable to download MLX Whisper model. The repository may be private "
                "or requires authentication. Provide a Hugging Face token via the "
                "HUGGINGFACEHUB_API_TOKEN / HF_TOKEN environment variable, or download "
                "the model manually and set 'mlx_model_dir' to the local path."
            ) from err
        except HfHubHTTPError as err:
            raise TranscriptionError(
                "Failed to contact Hugging Face Hub. Network access may be disabled on "
                "this environment. Download the MLX model manually and configure "
                "'mlx_model_dir' to point to it."
            ) from err
        except Exception as exc:  # pragma: no cover - defensive
            raise TranscriptionError(
                f"Unexpected error while downloading MLX model: {exc}"
            ) from exc


def _transcribe_local(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    if config.execution_mode == "internal":
        return _transcribe_local_internal(audio_path, config)
    elif config.execution_mode == "external":
        return _transcribe_local_external(audio_path, config)
    else:
        raise TranscriptionError(f"Unsupported execution mode for local backend: {config.execution_mode}")


def _transcribe_local_internal(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    try:
        import whisper  # type: ignore
    except ImportError as exc:
        raise TranscriptionError(
            "whisper Python package is required for internal transcription. "
            "Install it via 'pip install openai-whisper' or ensure the 'whisper' CLI is in your PATH."
        ) from exc

    device = config.device or None
    with _suppress_tqdm():
        model = whisper.load_model(config.model, device=device)
    transcribe_kwargs: Dict[str, Any] = dict(config.extra_args)
    if config.language:
        transcribe_kwargs.setdefault("language", config.language)
    if config.tune_segmentation:
        if config.temperature is not None:
            transcribe_kwargs.setdefault("temperature", config.temperature)
        if config.compression_ratio_threshold is not None:
            transcribe_kwargs.setdefault(
                "compression_ratio_threshold", config.compression_ratio_threshold
            )
        if config.logprob_threshold is not None:
            transcribe_kwargs.setdefault("logprob_threshold", config.logprob_threshold)
        if config.no_speech_threshold is not None:
            transcribe_kwargs.setdefault("no_speech_threshold", config.no_speech_threshold)
        if config.condition_on_previous_text is not None:
            transcribe_kwargs.setdefault(
                "condition_on_previous_text", config.condition_on_previous_text
            )
        if config.initial_prompt:
            transcribe_kwargs.setdefault("initial_prompt", config.initial_prompt)

    result = model.transcribe(str(audio_path), **transcribe_kwargs)
    segments = result.get("segments") or []
    if not segments:
        raise TranscriptionError("Whisper returned no segments")
    return SubtitleDocument.from_whisper_segments(segments)


def _transcribe_local_external(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    if not config.cli_path:
        raise TranscriptionError("cli_path not set for external local execution.")

    output_dir = audio_path.parent
    output_json = output_dir / f"{audio_path.stem}.json"

    cmd = [
        config.cli_path,
        str(audio_path),
        "--model",
        config.model,
        "--output_dir",
        str(output_dir),
        "--output_format",
        "json",
    ]
    if config.language:
        cmd.extend(["--language", config.language])
    if config.device:
        cmd.extend(["--device", config.device])

    # Add extra arguments from extra_args
    for key, value in (config.extra_args or {}).items():
        cmd.append(f"--{key}")
        if not isinstance(value, bool) or value:
            cmd.append(str(value))

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        raise TranscriptionError(
            f"Whisper CLI failed with exit code {result.returncode}: {result.stderr.strip()}"
        )

    if not output_json.exists():
        raise TranscriptionError(
            f"Whisper CLI did not produce JSON output at {output_json}"
        )

    with output_json.open("r", encoding="utf-8") as fh:
        payload = json.load(fh)
    
    output_json.unlink()

    segments = payload.get("segments") or []
    if not segments:
        raise TranscriptionError("Whisper CLI output did not contain segments")
    return SubtitleDocument.from_whisper_segments(segments)


def _transcribe_api(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    api_url = config.api_url or DEFAULT_OPENAI_TRANSCRIBE_URL
    api_key_env = config.api_key_env or "OPENAI_API_KEY"
    api_key = os.getenv(api_key_env)
    if not api_key:
        raise TranscriptionError(
            f"Environment variable {api_key_env} is not set for API transcription"
        )
    headers = {"Authorization": f"Bearer {api_key}"}
    data = {
        "model": config.model,
        **{k: v for k, v in (config.extra_args or {}).items()},
    }
    if config.language:
        data.setdefault("language", config.language)
    if config.tune_segmentation:
        if config.temperature is not None:
            data.setdefault("temperature", config.temperature)
        if config.initial_prompt:
            data.setdefault("prompt", config.initial_prompt)

    with audio_path.open("rb") as file_handle:
        files = {"file": (audio_path.name, file_handle)}
        response = requests.post(
            api_url, headers=headers, files=files, data=data, timeout=300
        )
    if response.status_code >= 400:
        raise TranscriptionError(
            f"Transcription API error {response.status_code}: {response.text.strip()}"
        )

    payload = response.json()
    segments = _extract_segments_from_api(payload)
    if not segments:
        raise TranscriptionError(
            "Transcription API returned unexpected payload without segments"
        )
    return SubtitleDocument.from_whisper_segments(segments)


def _transcribe_cpp(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    binary = config.cpp_binary or "whisper-cpp"
    exec_path = shutil.which(binary)
    if not exec_path:
        raise TranscriptionError(
            f"whisper.cpp executable '{binary}' not found in PATH."
        )
    if not config.cpp_model_path:
        raise TranscriptionError("cpp_model_path must be set for whisper.cpp backend.")

    output_prefix = audio_path.stem + "_whispercpp"
    output_json = audio_path.parent / f"{output_prefix}.json"
    try:
        output_json.unlink()
    except FileNotFoundError:
        pass
    except OSError:
        pass

    cmd: List[str] = [
        exec_path,
        "-f",
        str(audio_path),
        "-m",
        config.cpp_model_path,
        "-oj",
        "-of",
        str(audio_path.parent / output_prefix),
    ]
    if config.language:
        cmd.extend(["-l", config.language])
    if config.cpp_threads:
        cmd.extend(["-t", str(config.cpp_threads)])
    if config.cpp_extra_args:
        cmd.extend(config.cpp_extra_args)

    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        raise TranscriptionError(
            "whisper.cpp failed with exit code "
            f"{result.returncode}: {result.stderr.strip()}"
        )
    if not output_json.exists():
        raise TranscriptionError(
            f"whisper.cpp did not produce JSON output at {output_json}"
        )

    with output_json.open("r", encoding="utf-8") as fh:
        payload = json.load(fh)
    segments = _extract_segments_from_cpp(payload)
    try:
        output_json.unlink()
    except OSError:
        pass
    if not segments:
        raise TranscriptionError("whisper.cpp JSON output did not contain segments")
    return SubtitleDocument.from_whisper_segments(segments)


def _transcribe_mlx(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    if config.execution_mode == "internal":
        return _transcribe_mlx_internal(audio_path, config)
    elif config.execution_mode == "external":
        return _transcribe_mlx_external(audio_path, config)
    else:
        raise TranscriptionError(f"Unsupported execution mode for mlx backend: {config.execution_mode}")


def _transcribe_mlx_internal(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    try:
        import mlx_whisper  # type: ignore
    except ImportError as exc:
        raise TranscriptionError(
            "mlx-whisper package is required for internal execution. "
            "Install it or ensure the 'mlx_whisper' CLI tool is in your PATH for external execution."
        ) from exc

    kwargs: Dict[str, Any] = dict(config.mlx_extra_args or {})
    if config.language:
        kwargs.setdefault("language", config.language)
    if config.tune_segmentation:
        if config.no_speech_threshold is not None:
            kwargs.setdefault("no_speech_threshold", config.no_speech_threshold)
        if config.temperature is not None:
            kwargs.setdefault("temperature", config.temperature)
        if config.initial_prompt:
            kwargs.setdefault("initial_prompt", config.initial_prompt)

    model_identifier: str | None = None
    if config.mlx_model_dir:
        candidate_dir = Path(config.mlx_model_dir).expanduser()
        if not candidate_dir.exists():
            raise TranscriptionError(
                f"mlx_model_dir does not exist: {candidate_dir}"
            )
        model_identifier = str(candidate_dir)
    elif config.model:
        model_identifier = config.model.strip()

    if not model_identifier:
        repo_id = "mlx-community/whisper-small.en"
    else:
        candidate_path = Path(model_identifier)
        if candidate_path.exists():
            repo_id = str(candidate_path)
        elif model_identifier.startswith("mlx-community/") or "/" in model_identifier:
            repo_id = model_identifier
        else:
            repo_id = f"mlx-community/whisper-{model_identifier}"

    local_repo_path = _ensure_local_mlx_repo(repo_id)

    transcribe_fn = getattr(mlx_whisper, "transcribe", None)
    if not callable(transcribe_fn):
        raise TranscriptionError(
            "Installed mlx-whisper package does not expose a 'transcribe' function."
        )

    transcribe_kwargs = dict(kwargs)
    if config.mlx_dtype:
        dtype_lower = config.mlx_dtype.lower()
        if dtype_lower in {"float32", "fp32"}:
            transcribe_kwargs.setdefault("fp16", False)
        elif dtype_lower in {"float16", "fp16"}:
            transcribe_kwargs.setdefault("fp16", True)
    if config.mlx_device:
        transcribe_kwargs.setdefault("device", config.mlx_device)

    try:
        result = transcribe_fn(
            str(audio_path),
            path_or_hf_repo=local_repo_path,
            **transcribe_kwargs,
        )
    except TypeError:
        result = transcribe_fn(
            audio=str(audio_path),
            path_or_hf_repo=local_repo_path,
            **transcribe_kwargs,
        )

    if not isinstance(result, dict):
        raise TranscriptionError("mlx-whisper returned unexpected result format.")
    segments = result.get("segments") or []
    if not segments:
        raise TranscriptionError("mlx-whisper returned no segments.")
    return SubtitleDocument.from_whisper_segments(segments)


def _transcribe_mlx_external(audio_path: Path, config: WhisperConfig) -> SubtitleDocument:
    if not config.cli_path:
        raise TranscriptionError("cli_path not set for external mlx execution.")

    output_dir = Path(config.mlx_model_dir or '.')
    output_dir.mkdir(parents=True, exist_ok=True)
    output_name = audio_path.stem
    output_json = output_dir / f"{output_name}.json"

    cmd = [
        config.cli_path,
        str(audio_path),
        "--model",
        config.model,
        "--output-dir",
        str(output_dir),
        "--output-name",
        output_name,
        "--output-format",
        "json",
    ]
    if config.language:
        cmd.extend(["--language", config.language])

    # Add extra arguments from mlx_extra_args
    for key, value in (config.mlx_extra_args or {}).items():
        cmd.append(f"--{key}")
        if not isinstance(value, bool) or value:
            cmd.append(str(value))

    result = subprocess.run(cmd, capture_output=True, text=True, check=False)
    if result.returncode != 0:
        raise TranscriptionError(
            f"mlx_whisper CLI failed with exit code {result.returncode}: {result.stderr.strip()}"
        )

    if not output_json.exists():
        raise TranscriptionError(
            f"mlx_whisper CLI did not produce JSON output at {output_json}"
        )

    with output_json.open("r", encoding="utf-8") as fh:
        payload = json.load(fh)
    
    output_json.unlink()

    segments = payload.get("segments") or []
    if not segments:
        raise TranscriptionError("mlx_whisper CLI output did not contain segments")
    return SubtitleDocument.from_whisper_segments(segments)


def _extract_segments_from_api(payload: Dict[str, Any]) -> list[dict[str, Any]]:
    if "segments" in payload:
        return payload["segments"]
    if "text" in payload and "timestamps" in payload:
        return payload["timestamps"]
    if "text" in payload:
        return [
            {
                "start": 0,
                "end": 0,
                "text": payload["text"],
            }
        ]
    return []


def _extract_segments_from_cpp(payload: Dict[str, Any]) -> list[dict[str, Any]]:
    segments = payload.get("segments")
    if not segments and "transcription" in payload:
        segments = payload["transcription"].get("segments")
    results: list[dict[str, Any]] = []
    for segment in segments or []:
        text = segment.get("text")
        if not text:
            continue
        if "start" in segment and "end" in segment:
            start = float(segment["start"])
            end = float(segment["end"])
        else:
            t0 = segment.get("t0")
            t1 = segment.get("t1")
            if t0 is None or t1 is None:
                continue
            # whisper.cpp reports t0/t1 in 10ms units
            start = float(t0) * 0.01
            end = float(t1) * 0.01
        results.append(
            {
                "start": start,
                "end": end,
                "text": text,
            }
        )
    return results


__all__ = ["transcribe_audio", "TranscriptionError"]



from __future__ import annotations

import json
import os
import re
import time
from dataclasses import dataclass
from string import Template
from typing import Callable, Dict, Iterable, List, Optional

import requests

from .config import LLMConfig, PipelineConfig
from .subtitles import SubtitleDocument, SubtitleLine

DEFAULT_OPENAI_CHAT_URL = "https://api.openai.com/v1/chat/completions"


class LLMTranslationError(RuntimeError):
    """Raised when the LLM translation step fails."""


@dataclass
class TranslationChunk:
    index: int
    lines: List[SubtitleLine]


class LLMTranslator:
    def __init__(self, config: LLMConfig, pipeline: PipelineConfig) -> None:
        self.config = config
        self.pipeline = pipeline
        if config.api_base:
            base = config.api_base.rstrip("/")
            if base.endswith("chat/completions"):
                self.endpoint = base
            else:
                self.endpoint = f"{base}/chat/completions"
        else:
            self.endpoint = DEFAULT_OPENAI_CHAT_URL
        api_key = os.getenv(config.api_key_env)
        if not api_key:
            raise LLMTranslationError(
                f"Environment variable {config.api_key_env} is not set for LLM translation"
            )
        self.api_key = api_key
        self.total_prompt_tokens = 0
        self.total_completion_tokens = 0
        self.total_tokens = 0

    def translate_document(
        self,
        document: SubtitleDocument,
        existing_translations: Optional[Dict[str, str]] = None,
        progress_callback: Optional[Callable[[Dict[str, str]], None]] = None,
    ) -> SubtitleDocument:
        translations: Dict[str, str] = {
            str(int(key)): value.strip()
            for key, value in (existing_translations or {}).items()
        }
        for chunk_index, chunk in enumerate(self._iter_chunks(document), start=1):
            if all(str(line.index) in translations for line in chunk.lines):
                continue
            payload = self._build_payload(chunk_index, chunk)
            attempt = 0
            while True:
                try:
                    response = self._invoke(payload)
                    chunk_translations = self._parse_response(response, chunk)
                    self._update_usage(response.get("usage") or {})
                    update_payload: Dict[str, str] = {}
                    for line, translation in zip(
                        chunk.lines, chunk_translations, strict=True
                    ):
                        translations[str(line.index)] = translation
                        update_payload[str(line.index)] = translation
                    if progress_callback and update_payload:
                        progress_callback(update_payload)
                    break
                except (requests.RequestException, json.JSONDecodeError, LLMTranslationError) as exc:
                    attempt += 1
                    if attempt > self.config.max_retries:
                        raise LLMTranslationError(
                            f"Failed to translate chunk {chunk_index}: {exc}"
                        ) from exc
                    time.sleep(2**attempt * 0.5)
        final_lines: List[SubtitleLine] = []
        for line in document.lines:
            translated_text = translations.get(str(line.index))
            if translated_text is None:
                raise LLMTranslationError(
                    f"Missing translation for line index {line.index}."
                )
            final_lines.append(
                SubtitleLine(
                    index=line.index,
                    start=line.start,
                    end=line.end,
                    text=translated_text,
                )
            )
        self.total_tokens = self.total_prompt_tokens + self.total_completion_tokens
        return SubtitleDocument(lines=final_lines)

    def token_usage(self) -> Dict[str, int]:
        return {
            "prompt": self.total_prompt_tokens,
            "completion": self.total_completion_tokens,
            "total": self.total_tokens,
        }

    def _iter_chunks(self, document: SubtitleDocument) -> Iterable[TranslationChunk]:
        for idx, lines in enumerate(document.chunk(self.config.batch_size), start=1):
            yield TranslationChunk(index=idx, lines=list(lines))

    def _build_payload(self, chunk_index: int, chunk: TranslationChunk) -> dict:
        user_prompt = self._format_prompt(chunk)
        system_prompt = Template(self.pipeline.prompt_preamble).safe_substitute(
            targetLanguage=self.config.target_language,
            style=self.config.style or "",
            model=self.config.model,
            provider=self.config.provider,
        )
        messages = [
            {
                "role": "system",
                "content": system_prompt,
            },
            {
                "role": "user",
                "content": user_prompt,
            },
        ]
        return {
            "model": self.config.model,
            "messages": messages,
            "temperature": self.config.temperature,
        }

    def _invoke(self, payload: dict) -> dict:
        headers = {
            "Authorization": f"Bearer {self.api_key}",
            "Content-Type": "application/json",
        }
        response = requests.post(
            self.endpoint,
            headers=headers,
            json=payload,
            timeout=self.config.request_timeout,
        )
        if response.status_code >= 400:
            raise LLMTranslationError(
                f"LLM API error {response.status_code}: {response.text.strip()}"
            )
        try:
            return response.json()
        except ValueError as exc:
            snippet = response.text.strip().splitlines()
            preview = snippet[0][:200] if snippet else ""
            raise LLMTranslationError(
                "LLM API returned non-JSON payload. Partial response: "
                f"{preview}"
            ) from exc

    def _parse_response(self, response: dict, chunk: TranslationChunk) -> List[str]:
        choices = response.get("choices")
        if not choices:
            raise LLMTranslationError("LLM response missing choices")
        message = choices[0].get("message")
        if not message:
            raise LLMTranslationError("LLM response missing message content")
        content = message.get("content", "").strip()
        parsed = self._load_json_response(content)
        if not isinstance(parsed, dict):
            raise LLMTranslationError("Translation response must be a JSON object.")

        expected_keys = [str(line.index) for line in chunk.lines]
        missing_keys = [key for key in expected_keys if key not in parsed]
        extra_keys = [key for key in parsed.keys() if key not in expected_keys]
        if missing_keys:
            raise LLMTranslationError(
                f"Missing translations for keys: {', '.join(missing_keys)}"
            )
        if extra_keys:
            raise LLMTranslationError(
                f"Received unexpected keys in translation: {', '.join(extra_keys)}"
            )

        cleaned: List[str] = []
        for key in expected_keys:
            value = parsed[key]
            if not isinstance(value, str):
                raise LLMTranslationError(
                    f"Translation value for key {key} is not a string"
                )
            cleaned.append(value.strip())
        return cleaned

    def _load_json_response(self, content: str) -> dict:
        """Attempt to parse JSON content, tolerating Markdown fencing."""

        try:
            return json.loads(content)
        except json.JSONDecodeError:
            pass

        fenced_match = re.search(r"```(?:json)?\s*(\{.*?\})\s*```", content, re.DOTALL)
        if fenced_match:
            fenced_content = fenced_match.group(1)
            try:
                return json.loads(fenced_content)
            except json.JSONDecodeError:
                pass

        start = content.find("{")
        end = content.rfind("}")
        if start != -1 and end != -1 and end > start:
            candidate = content[start : end + 1]
            try:
                return json.loads(candidate)
            except json.JSONDecodeError:
                pass

        raise LLMTranslationError(
            "LLM response was not valid JSON. Ensure the model returns a pure JSON object."
        )

    def _update_usage(self, usage: Dict[str, object]) -> None:
        prompt = self._coerce_int(
            usage.get("prompt_tokens")
            or usage.get("prompt_token")
            or usage.get("prompt")
        )
        completion = self._coerce_int(
            usage.get("completion_tokens")
            or usage.get("completion_token")
            or usage.get("completion")
        )
        total = self._coerce_int(usage.get("total_tokens") or usage.get("total"))

        if prompt is not None:
            self.total_prompt_tokens += prompt
        if completion is not None:
            self.total_completion_tokens += completion
        if total is not None:
            self.total_tokens = max(self.total_tokens, total)

    @staticmethod
    def _coerce_int(value: object) -> Optional[int]:
        if value is None:
            return None
        try:
            return int(value)
        except (TypeError, ValueError):
            return None

    def _format_prompt(self, chunk: TranslationChunk) -> str:
        batch_payload = {str(line.index): line.text for line in chunk.lines}
        batch_json = json.dumps(
            batch_payload,
            ensure_ascii=False,
            separators=(",", ":"),
        )
        style_instruction = (
            f"Target style: {self.config.style}" if self.config.style else "Maintain a conversational tone."
        )
        return (
            "The following JSON object contains subtitle lines to translate; each key maps to one line.\n"
            f"Target language: {self.config.target_language}.\n"
            f"{style_instruction}\n"
            "Translate only the values, keep the original keys, do not add or remove entries, and do not surround the result with explanations.\n"
            "Return a valid JSON object that can be parsed directly.\n"
            "Content to translate:\n"
            f"{batch_json}\n"
            "Respond with the translated JSON now."
        )


def translate_subtitles(
    document: SubtitleDocument,
    config: LLMConfig,
    pipeline: PipelineConfig,
    existing_translations: Optional[Dict[str, str]] = None,
    progress_callback: Optional[Callable[[Dict[str, str]], None]] = None,
) -> tuple[SubtitleDocument, Dict[str, int]]:
    translator = LLMTranslator(config=config, pipeline=pipeline)
    doc = translator.translate_document(
        document,
        existing_translations=existing_translations,
        progress_callback=progress_callback,
    )
    usage = translator.token_usage()
    return doc, usage


__all__ = ["translate_subtitles", "LLMTranslationError"]



from __future__ import annotations

import json
import shlex
from pathlib import Path
from dataclasses import dataclass
from typing import Callable, Dict, List, Optional
from importlib import metadata

import typer
from rich import box
from rich.console import Console, Group, RenderableType
from rich.panel import Panel
from rich.prompt import Confirm, Prompt
from rich.progress import SpinnerColumn, Progress, TextColumn, TaskProgressColumn
from rich.padding import Padding
from rich.table import Table
from rich.text import Text
from .audio import AudioExtractionError, extract_audio
from .config import ConfigManager, TransubConfig, DEFAULT_TRANSLATION_PROMPT
from .logger import setup_logging
from .state import (
    PipelineState,
    load_translation_progress,
    persist_translation_progress,
)
from .subtitles import SubtitleDocument
from .transcribe import TranscriptionError, transcribe_audio, check_dependencies
from .translate import LLMTranslationError, translate_subtitles

THEME_COLOR = "#33c9b2"
THEME_VARIABLE_COLOR = "#f0b429"
THEME_VARIABLE_STYLE = f"bold {THEME_VARIABLE_COLOR}"
THEME_HIGHLIGHT = f"bold {THEME_COLOR}"

app = typer.Typer(add_completion=False, help="Transcribe and translate subtitles from videos.")
console = Console()
WHISPER_MODEL_SUGGESTIONS: dict[str, list[str]] = {
    "local": [
        "small",
        "medium",
        "large-v3",
        "large-v2",
        "base",
        "tiny",
    ],
    "mlx": [
        "mlx-community/whisper-small.en-mlx",
        "mlx-community/whisper-medium.en-mlx",
        "mlx-community/whisper-large-v3",
        "mlx-community/whisper-large-v2",
    ],
    "api": [
        "gpt-4o-mini-transcribe",
        "gpt-4o-transcribe",
        "whisper-1",
    ],
    "cpp": [
        "ggml-small.en.bin",
        "ggml-medium.en.bin",
        "ggml-large-v3.bin",
        "gguf-large-v3-q5_1.bin",
    ],
}


def _get_version() -> str:
    try:
        return metadata.version("transub")
    except metadata.PackageNotFoundError:  # pragma: no cover - dev installs
        return "unknown"


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    version: bool = typer.Option(
        False,
        "--version",
        "-v",
        help="Show Transub version and exit.",
        is_eager=True,
    ),
) -> None:
    if version:
        console.print(f"Transub { _get_version() }")
        raise typer.Exit()
    if ctx.invoked_subcommand is None:
        console.print(app.get_help())
def _print_header(
    *,
    subtitle: str | None = None,
    video: Path | None = None,
    body: RenderableType | None = None,
) -> None:
    header = Table.grid(expand=True)
    header.add_column(ratio=1)
    header.add_column(justify="right")

    title = Text("Transub", style=THEME_HIGHLIGHT)
    if subtitle:
        title.append("  ·  ", style="dim")
        title.append(subtitle, style="white")
    tagline = Text("video → subtitles (transcribe + translate)", style="bright_black")

    header.add_row(title, tagline)

    renderables: List[RenderableType] = [header]

    if video:
        video_table = _key_value_table([("source", str(video))])
        renderables.append(video_table)

    if body is not None:
        renderables.append(body)

    panel = Panel(
        Group(*renderables),
        border_style=THEME_COLOR,
        padding=(1, 2),
    )
    console.print(panel)


@app.command()
def init(
    config_path: Optional[Path] = typer.Option(
        None,
        "--config",
        "-c",
        help="Custom path for the configuration file",
    )
) -> None:
    """Guided configuration setup."""

    manager = ConfigManager(config_path or ConfigManager.default_path())
    _print_header(subtitle="configuration wizard")
    _run_wizard(manager, allow_overwrite=True)


@app.command(context_settings={"allow_extra_args": True, "ignore_unknown_options": True})
def run(
    video: Path = typer.Argument(..., exists=True, readable=True, help="Path to the video file"),
    config_path: Optional[Path] = typer.Option(
        None, "--config", "-c", help="Custom configuration file location"
    ),
    work_dir: Optional[Path] = typer.Option(
        None,
        "--work-dir",
        help="Working directory for intermediate files (defaults to ~/.cache/transub).",
    ),
    transcribe_only: bool = typer.Option(
        False,
        "--transcribe-only",
        "-T",
        help="Skip translation and export the transcription only.",
    ),
) -> None:
    """Run the end-to-end subtitle creation pipeline."""

    config = _load_config(config_path)
    check_dependencies(config.whisper)

    work_dir = (work_dir or Path.home() / ".cache" / "transub").resolve()
    work_dir.mkdir(parents=True, exist_ok=True)

    config_display = (config_path or ConfigManager.default_path()).resolve()
    mode_label = "transcribe-only" if transcribe_only else "full pipeline"
    source_lang = config.whisper.language or "auto"
    target_lang = config.llm.target_language or "auto"
    source_text = Text(source_lang, style=THEME_VARIABLE_STYLE)
    target_text = Text(target_lang, style=THEME_VARIABLE_STYLE)
    run_snapshot = _key_value_table(
        [
            ("config", str(config_display)),
            ("work_dir", str(work_dir)),
            ("mode", mode_label),
            ("source language", source_text),
            ("target language", target_text),
        ]
    )

    _print_header(subtitle="pipeline", video=video, body=run_snapshot)

    logger = setup_logging(work_dir / "transub.log")
    logger.info("Starting pipeline for %s", video)

    state_path = work_dir / f"{video.stem}_state.json"
    state = PipelineState.load(state_path, video)

    audio_path: Optional[Path] = state.get_audio_path()
    segments_path: Optional[Path] = state.get_segments_path()
    translations_path: Optional[Path] = None
    interrupted = False

    success = False
    try:
        if audio_path and audio_path.exists():
            audio_path = audio_path.resolve()
            console.print(f"Using cached audio at [italic]{audio_path}[/]")
            logger.info("Using cached audio %s", audio_path)
        else:
            with console.status("Extracting audio…", spinner="dots"):
                audio_path = extract_audio(video, config.pipeline, work_dir)
            audio_path = audio_path.resolve()
            state.set_audio_path(audio_path)
            console.print(f"Audio extracted to [italic]{audio_path}[/]")
            logger.info("Audio extracted to %s", audio_path)

        if segments_path and segments_path.exists():
            with segments_path.open("r", encoding="utf-8") as fh:
                segment_payload = json.load(fh)
            source_doc = SubtitleDocument.from_serialized(segment_payload)
            console.print("Loaded cached transcription.")
            logger.info(
                "Loaded cached transcription from %s (%d lines)",
                segments_path,
                len(source_doc.lines),
            )
            if state.transcription_total_lines() is None:
                state.mark_transcription(segments_path, len(source_doc.lines))
        else:
            with console.status(
                "Transcribing audio with Whisper…", spinner="dots"
            ):
                raw_doc = transcribe_audio(audio_path, config.whisper)
            refined_doc = raw_doc.refine(
                max_chars=config.pipeline.max_chars_per_line,
                min_chars=config.pipeline.min_chars_per_line,
            )
            if (
                config.pipeline.timing_trim_seconds > 0
                or config.pipeline.timing_offset_seconds != 0
            ):
                refined_doc = refined_doc.adjust_timing(
                    trim=config.pipeline.timing_trim_seconds,
                    min_duration=config.pipeline.min_line_duration,
                    offset=config.pipeline.timing_offset_seconds,
                )
            segments_path = work_dir / f"{video.stem}_segments.json"
            segments_path.write_text(
                json.dumps(refined_doc.to_serializable(), ensure_ascii=False, indent=2),
                encoding="utf-8",
            )
            state.mark_transcription(segments_path, len(refined_doc.lines))
            source_doc = refined_doc
            console.print("Transcription complete.")
            logger.info(
                "Transcription finished with %d refined lines", len(refined_doc.lines)
            )

        total_lines = state.transcription_total_lines() or len(source_doc.lines)
        if total_lines <= 0:
            total_lines = len(source_doc.lines)
            if segments_path:
                state.mark_transcription(segments_path, total_lines)

        output_dir = (
            Path(config.pipeline.output_dir)
            if config.pipeline.output_dir is not None
            else video.parent
        )

        if transcribe_only:
            console.print(f"[{THEME_COLOR}]Transcribe-only mode: skipping translation.[/]")
            logger.info("Transcribe-only mode enabled; skipping translation stage.")

            transcript_doc = source_doc
            max_trans_chars = config.pipeline.translation_max_chars_per_line
            min_trans_chars: int | None = None
            if (
                config.pipeline.refine_source_subtitles
                and max_trans_chars
            ):
                min_trans_chars = (
                    config.pipeline.translation_min_chars_per_line
                    or min(config.pipeline.min_chars_per_line, max_trans_chars)
                )
                transcript_doc = source_doc.refine(
                    max_chars=max_trans_chars,
                    min_chars=min_trans_chars,
                )

            source_suffix = _source_language_suffix(config.whisper.language)
            transcript_path = _write_document(
                document=transcript_doc,
                target_dir=output_dir,
                stem=video.stem,
                suffix=source_suffix,
                output_format=config.pipeline.output_format,
            )
            console.print(Panel.fit(f"✅ Transcription saved to {transcript_path}", style="green"))
            logger.info("Transcription exported to %s", transcript_path)

            success = True
            return

        translations_path = state.translation_progress_path(
            work_dir / f"{video.stem}_translations.json"
        )
        existing_translations: Dict[str, str] = load_translation_progress(translations_path)
        if existing_translations:
            state.mark_lines_completed(existing_translations.keys())
            logger.info(
                "Loaded cached translations for %d lines",
                len(existing_translations),
            )
            console.print(
                f"Resuming translation: {len(existing_translations)}/{total_lines} lines already completed."
            )

        translations_cache: Dict[str, str] = dict(existing_translations)

        initial_completed = len(translations_cache)

        def _progress_description(done: int) -> str:
            return f"[{THEME_HIGHLIGHT}]Translating[/] {done}/{total_lines}"

        progress = Progress(
            SpinnerColumn(style=THEME_COLOR),
            TextColumn("{task.description}"),
            TaskProgressColumn(),
            console=console,
            transient=True,
        )

        with progress:
            task_id = progress.add_task(
                description=_progress_description(initial_completed),
                total=total_lines,
                completed=initial_completed,
            )

            def handle_progress(new_items: Dict[str, str]) -> None:
                translations_cache.update(new_items)
                persist_translation_progress(translations_path, translations_cache)
                state.mark_lines_completed(new_items.keys())
                done = len(translations_cache)
                progress.update(
                    task_id,
                    completed=done,
                    description=_progress_description(done),
                )
                logger.info(
                    "Translated lines %s",
                    ", ".join(sorted(new_items.keys(), key=int)),
                )

            translated_doc, usage_stats = translate_subtitles(
                source_doc,
                config.llm,
                config.pipeline,
                existing_translations=translations_cache,
                progress_callback=handle_progress,
            )
        console.print(
            "Translation complete. Tokens used: "
            f"prompt {usage_stats['prompt']}, "
            f"completion {usage_stats['completion']}, "
            f"total {usage_stats['total']}"
            f" (translated {len(translations_cache)}/{total_lines} lines)"
        )
        persist_translation_progress(translations_path, translations_cache)

        output_doc = translated_doc
        max_trans_chars = config.pipeline.translation_max_chars_per_line
        min_trans_chars = None
        if max_trans_chars:
            min_trans_chars = (
                config.pipeline.translation_min_chars_per_line
                or min(config.pipeline.min_chars_per_line, max_trans_chars)
            )
            output_doc = translated_doc.refine(
                max_chars=max_trans_chars,
                min_chars=min_trans_chars,
            )

        if config.pipeline.remove_trailing_punctuation:
            output_doc = output_doc.remove_trailing_punctuation()
        if config.pipeline.normalize_cjk_spacing:
            output_doc = output_doc.normalize_cjk_spacing()

        output_suffix = _language_suffix(config.llm.target_language)
        output_path = _write_document(
            document=output_doc,
            target_dir=output_dir,
            stem=video.stem,
            suffix=output_suffix,
            output_format=config.pipeline.output_format,
        )
        console.print(Panel.fit(f"✅ Finished! Output saved to {output_path}", style="green"))
        logger.info(
            "Translation finished. Output saved to %s. Tokens prompt=%d completion=%d total=%d",
            output_path,
            usage_stats["prompt"],
            usage_stats["completion"],
            usage_stats["total"],
        )

        if config.pipeline.save_source_subtitles:
            source_output_doc = source_doc
            if (
                config.pipeline.refine_source_subtitles
                and max_trans_chars
                and min_trans_chars is not None
            ):
                source_output_doc = source_doc.refine(
                    max_chars=max_trans_chars,
                    min_chars=min_trans_chars,
                )
            source_suffix = _source_language_suffix(config.whisper.language)
            source_path = _write_document(
                document=source_output_doc,
                target_dir=output_dir,
                stem=video.stem,
                suffix=source_suffix,
                output_format=config.pipeline.output_format,
            )
            console.print(f"Source subtitles saved to [italic]{source_path}[/]")
            logger.info("Source subtitles saved to %s", source_path)

        success = True

    except KeyboardInterrupt as exc:  # pragma: no cover - manual interrupt
        interrupted = True
        logger.info("Pipeline interrupted by user.")
        console.print("[yellow]Pipeline interrupted by user. Partial results kept.[/]")
        raise typer.Exit(code=1) from exc
    except FileNotFoundError as exc:
        logger.exception("Missing dependency: %s", exc)
        console.print(f"[red]Missing dependency:[/] {exc}")
        raise typer.Exit(code=1) from exc
    except AudioExtractionError as exc:
        logger.exception("Audio extraction failed: %s", exc)
        console.print(f"[red]Audio extraction failed:[/] {exc}")
        raise typer.Exit(code=1) from exc
    except TranscriptionError as exc:
        logger.exception("Transcription failed: %s", exc)
        console.print(f"[red]Transcription failed:[/] {exc}")
        raise typer.Exit(code=1) from exc
    except LLMTranslationError as exc:
        logger.exception("Translation failed: %s", exc)
        console.print(f"[red]Translation failed:[/] {exc}")
        raise typer.Exit(code=1) from exc
    finally:
        if success:
            if (
                config
                and work_dir.exists()
                and not config.pipeline.keep_temp_audio
                and audio_path
            ):
                _cleanup_audio_file(work_dir, audio_path)
            try:
                if translations_path and translations_path.exists():
                    translations_path.unlink()
            except OSError:
                pass
            try:
                current_segments = state.get_segments_path()
                if current_segments and current_segments.exists():
                    current_segments.unlink()
            except OSError:
                pass
            state.clear()
        else:
            _offer_failure_cleanup(
                work_dir=work_dir,
                state=state,
                audio_path=audio_path,
                translations_path=translations_path,
                interrupted=interrupted,
            )


def _load_config(config_path: Optional[Path]) -> TransubConfig:
    manager = ConfigManager(config_path or ConfigManager.default_path())
    if manager.exists():
        return manager.load()

    target_path = manager.path
    console.print(
        Panel.fit(
            f"No configuration found at {target_path}.",
            style="yellow",
        )
    )
    if Confirm.ask("Run the setup wizard now?", default=True):
        _run_wizard(manager, allow_overwrite=True)
        return manager.load()

    console.print(
        "[red]Configuration missing.[/] Run `transub init` or specify --config pointing to an existing file."
    )
    raise typer.Exit(code=1)


@app.command()
def show_config(
    config_path: Optional[Path] = typer.Option(
        None, "--config", "-c", help="Custom configuration file location"
    )
) -> None:
    """Display the current configuration."""

    manager = ConfigManager(config_path or ConfigManager.default_path())
    if not manager.exists():
        console.print("Configuration file not found. Run `transub init` first.")
        raise typer.Exit(code=1)
    config = manager.load()
    console.print(json.dumps(config.model_dump(mode="json"), indent=2, ensure_ascii=False))


@app.command()
def configure(
    config_path: Optional[Path] = typer.Option(
        None, "--config", "-c", help="Custom configuration file location"
    )
) -> None:
    """Interactively adjust existing configuration."""

    manager = ConfigManager(config_path or ConfigManager.default_path())
    if not manager.exists():
        console.print("Configuration file not found. Run `transub init` first.")
        raise typer.Exit(code=1)

    config = manager.load()
    option_handlers = [
        ("Whisper backend & model", _configure_whisper_backend),
        ("Whisper advanced parameters", _configure_whisper_advanced),
        ("Translation LLM", _configure_llm),
        ("Pipeline & output", _configure_pipeline),
        ("View raw JSON", None),
    ]

    dirty = False
    while True:
        console.clear()
        _print_header(
            subtitle="configuration editor",
            body=_config_summary_table(config, manager.path),
        )
        console.print()

        console.print("[bold]Select an option:[/]")
        menu = Table(
            show_header=False,
            box=box.SIMPLE_HEAD,
            expand=False,
            padding=(0, 1),
        )
        menu.add_column(style=THEME_HIGHLIGHT, justify="center", width=4)
        menu.add_column(style="white", justify="left")
        menu.add_row("0", "Save & exit")
        menu.add_row("Q", "Exit without saving")
        for idx, (label, _) in enumerate(option_handlers, start=1):
            menu.add_row(str(idx), label)
        console.print(menu)

        choice = Prompt.ask("Your choice", default="0").strip()
        if not choice:
            choice = "0"

        if choice.lower() in {"q", "quit", "exit"}:
            if dirty and not Confirm.ask("Discard unsaved changes?", default=False):
                continue
            console.clear()
            console.print(Panel.fit("Changes discarded.", style="yellow"))
            return

        if choice == "0":
            manager.save(config)
            dirty = False
            console.clear()
            console.print(
                Panel.fit(f"Configuration saved to {manager.path}", style="green")
            )
            return

        try:
            idx = int(choice)
        except ValueError:
            _wait_for_enter("Please enter a valid number. Press Enter to continue...")
            continue
        if idx < 1 or idx > len(option_handlers):
            _wait_for_enter("Selection out of range. Press Enter to continue...")
            continue
        label, handler = option_handlers[idx - 1]
        if handler is None:
            console.clear()
            console.print(
                Panel(
                    json.dumps(
                        config.model_dump(mode="json"),
                        indent=2,
                        ensure_ascii=False,
                    ),
                    title="Raw configuration",
                )
            )
            _wait_for_enter()
        else:
            handler(config)
            dirty = True


BACK_TOKENS = {"back", "b"}


class BackAction(Exception):
    """Signal that the wizard should move back one step."""


@dataclass
class WizardStep:
    key: str
    heading: str
    description: Callable[[TransubConfig], Optional[str]]
    handler: Callable[[TransubConfig], None]
    condition: Callable[[TransubConfig], bool] = lambda _: True


def _is_back(value: str) -> bool:
    return value.strip().lower() in BACK_TOKENS



def _wizard_step_body(title: str, description: Optional[str]) -> RenderableType:
    lines: List[RenderableType] = [Text(title, style="bold white")]
    if description:
        lines.append(Text(description, style="white"))
    lines.append(Text("Type 'back' to return to the previous step.", style="dim"))
    return Padding(Group(*lines), (0, 0, 0, 0))


def _wizard_ask_choice(prompt_text: str, choices: List[str], default: str) -> str:
    choice_display = "/".join(choices)
    normalized = [choice.lower() for choice in choices]
    while True:
        raw = Prompt.ask(
            f"{prompt_text} ({choice_display})",
            default=default,
            show_choices=False,
        )
        if raw is None:
            raw = ""
        if _is_back(raw):
            raise BackAction()
        value = raw.strip()
        if value.lower() in normalized:
            return choices[normalized.index(value.lower())]
        console.print(
            "[red]Please choose one of the listed options, or type 'back' to return.[/]"
        )


def _wizard_ask_text(
    prompt_text: str,
    *,
    default: str | None = None,
    allow_blank: bool = True,
    to_none: bool = False,
) -> Optional[str]:
    while True:
        raw = Prompt.ask(
            f"{prompt_text}",
            default=default,
            show_choices=False,
        )
        if raw is None:
            raw = ""
        if _is_back(raw):
            raise BackAction()
        value = raw.strip()
        if not value and not allow_blank:
            console.print("[red]This field cannot be empty.[/]")
            continue
        if to_none:
            return value or None
        return value if value or allow_blank else default


def _wizard_ask_bool(prompt_text: str, default: bool) -> bool:
    default_hint = "Y/n" if default else "y/N"
    default_value = "y" if default else "n"
    while True:
        raw = Prompt.ask(
            f"{prompt_text} ({default_hint})",
            default=default_value,
            show_choices=False,
        )
        if raw is None:
            raw = default_value
        if _is_back(raw):
            raise BackAction()
        value = raw.strip().lower()
        if value in {"y", "yes"}:
            return True
        if value in {"n", "no"}:
            return False
        if not value:
            return default
        console.print("[red]Please enter y or n.[/]")


def _wizard_ask_int(
    prompt_text: str,
    *,
    default: int,
    minimum: int,
    maximum: int,
) -> int:
    while True:
        raw = Prompt.ask(
            f"{prompt_text} [{minimum}-{maximum}]",
            default=str(default),
            show_choices=False,
        )
        if raw is None:
            raw = str(default)
        if _is_back(raw):
            raise BackAction()
        try:
            value = int(raw)
        except ValueError:
            console.print("[red]Please enter an integer value.[/]")
            continue
        if value < minimum or value > maximum:
            console.print(
                f"[red]Value must be between {minimum} and {maximum}.[/]"
            )
            continue
        return value


def _wizard_ask_optional_int(
    prompt_text: str,
    *,
    current: Optional[int],
    minimum: int,
) -> Optional[int]:
    default = "" if current is None else str(current)
    while True:
        raw = Prompt.ask(
            f"{prompt_text} (leave blank for auto)",
            default=default,
            show_choices=False,
        )
        if raw is None:
            raw = ""
        if _is_back(raw):
            raise BackAction()
        value = raw.strip()
        if not value:
            return None
        try:
            parsed = int(value)
        except ValueError:
            console.print("[red]Please enter an integer or leave blank.[/]")
            continue
        if parsed < minimum:
            console.print(f"[red]Value must be at least {minimum}.[/]")
            continue
        return parsed


def _wizard_ask_json_dict(
    prompt_text: str,
    *,
    default: dict,
) -> dict:
    default_text = json.dumps(default or {}, ensure_ascii=False)
    while True:
        raw = Prompt.ask(
            f"{prompt_text} (JSON)",
            default=default_text,
            show_choices=False,
        )
        if raw is None:
            raw = default_text
        if _is_back(raw):
            raise BackAction()
        value = raw.strip()
        if not value:
            return {}
        try:
            parsed = json.loads(value)
        except json.JSONDecodeError:
            console.print("[red]Please enter a valid JSON object.[/]")
            continue
        if not isinstance(parsed, dict):
            console.print("[red]The value must be a JSON object (key/value pairs).[/]")
            continue
        return parsed


def _wizard_step_whisper_backend(config: TransubConfig) -> None:
    choices = ["local", "api", "cpp", "mlx"]
    backend = _wizard_ask_choice(
        "Select backend",
        choices,
        config.whisper.backend,
    )
    config.whisper.backend = backend


def _wizard_step_whisper_model(config: TransubConfig) -> None:
    suggestions = WHISPER_MODEL_SUGGESTIONS.get(config.whisper.backend, [])
    default_model = config.whisper.model or (suggestions[0] if suggestions else "base")
    model = _wizard_ask_text(
        "Model ID or path",
        default=default_model,
        allow_blank=False,
    )
    assert model
    config.whisper.model = model


def _wizard_step_whisper_device(config: TransubConfig) -> None:
    device = _wizard_ask_text(
        "Preferred device",
        default=config.whisper.device or "",
        to_none=True,
    )
    config.whisper.device = device


def _wizard_step_whisper_language(config: TransubConfig) -> None:
    language = _wizard_ask_text(
        "Source language",
        default=config.whisper.language or "en",
        to_none=True,
    )
    config.whisper.language = language


def _wizard_step_whisper_segmentation(config: TransubConfig) -> None:
    config.whisper.tune_segmentation = _wizard_ask_bool(
        "Enable segmentation tuning?",
        config.whisper.tune_segmentation,
    )


def _wizard_step_whisper_initial_prompt(config: TransubConfig) -> None:
    initial_prompt = _wizard_ask_text(
        "Initial prompt for Whisper (optional)",
        default=config.whisper.initial_prompt or "",
        to_none=True,
    )
    config.whisper.initial_prompt = initial_prompt


def _wizard_step_whisper_api(config: TransubConfig) -> None:
    default_url = config.whisper.api_url or DEFAULT_OPENAI_TRANSCRIBE_URL
    api_url = _wizard_ask_text(
        "API URL",
        default=default_url,
        allow_blank=False,
    )
    api_key_env = _wizard_ask_text(
        "API key environment variable",
        default=config.whisper.api_key_env,
        allow_blank=False,
    )
    config.whisper.api_url = api_url
    config.whisper.api_key_env = api_key_env


def _wizard_step_whisper_cpp(config: TransubConfig) -> None:
    config.whisper.cpp_binary = _wizard_ask_text(
        "whisper.cpp executable",
        default=config.whisper.cpp_binary,
        allow_blank=False,
    )
    config.whisper.cpp_model_path = _wizard_ask_text(
        "Model file path (.bin/.gguf)",
        default=config.whisper.cpp_model_path or "",
        allow_blank=False,
    )
    config.whisper.cpp_threads = _wizard_ask_optional_int(
        "Thread count",
        current=config.whisper.cpp_threads,
        minimum=1,
    )
    extra_args = _wizard_ask_text(
        "Additional arguments (space separated)",
        default=" ".join(config.whisper.cpp_extra_args) if config.whisper.cpp_extra_args else "",
    )
    config.whisper.cpp_extra_args = shlex.split(extra_args) if extra_args else []


def _wizard_step_whisper_mlx(config: TransubConfig) -> None:
    config.whisper.mlx_model_dir = _wizard_ask_text(
        "Model directory",
        default=config.whisper.mlx_model_dir or "",
        to_none=True,
    )
    config.whisper.mlx_dtype = _wizard_ask_text(
        "dtype (auto/float16/float32)",
        default=config.whisper.mlx_dtype or "",
        to_none=True,
    )
    config.whisper.mlx_device = _wizard_ask_text(
        "Device (auto/mps/cpu)",
        default=config.whisper.mlx_device or "",
        to_none=True,
    )
    extra_args = _wizard_ask_json_dict(
        "Extra arguments",
        default=config.whisper.mlx_extra_args or {},
    )
    config.whisper.mlx_extra_args = extra_args


def _wizard_step_llm_provider(config: TransubConfig) -> None:
    config.llm.provider = _wizard_ask_text(
        "LLM provider",
        default=config.llm.provider,
        allow_blank=False,
    )
    config.llm.model = _wizard_ask_text(
        "LLM model",
        default=config.llm.model,
        allow_blank=False,
    )
    api_base = _wizard_ask_text(
        "API base URL (leave blank for default)",
        default=config.llm.api_base or "",
        to_none=True,
    )
    config.llm.api_base = api_base
    config.llm.api_key_env = _wizard_ask_text(
        "API key environment variable",
        default=config.llm.api_key_env,
        allow_blank=False,
    )


def _wizard_step_llm_batch_and_style(config: TransubConfig) -> None:
    config.llm.batch_size = _wizard_ask_int(
        "Lines per translation batch",
        default=config.llm.batch_size,
        minimum=1,
        maximum=50,
    )
    config.llm.target_language = _wizard_ask_text(
        "Target language",
        default=config.llm.target_language,
        allow_blank=False,
    )
    style = _wizard_ask_text(
        "Style description (optional)",
        default=config.llm.style or "",
        to_none=True,
    )
    config.llm.style = style


def _wizard_step_pipeline_output(config: TransubConfig) -> None:
    config.pipeline.output_format = _wizard_ask_choice(
        "Subtitle format",
        ["srt", "vtt"],
        config.pipeline.output_format,
    )
    config.pipeline.audio_format = _wizard_ask_choice(
        "Intermediate audio format",
        ["wav", "mp3", "flac", "m4a", "ogg"],
        config.pipeline.audio_format,
    )
    config.pipeline.output_dir = _wizard_ask_text(
        "Output directory",
        default=config.pipeline.output_dir,
        allow_blank=False,
    )
    config.pipeline.keep_temp_audio = _wizard_ask_bool(
        "Keep extracted audio file?",
        config.pipeline.keep_temp_audio,
    )
    config.pipeline.save_source_subtitles = _wizard_ask_bool(
        "Save source language subtitles?",
        config.pipeline.save_source_subtitles,
    )


def _wizard_step_pipeline_limits(config: TransubConfig) -> None:
    config.pipeline.max_chars_per_line = _wizard_ask_int(
        "Maximum characters per source line",
        default=config.pipeline.max_chars_per_line,
        minimum=20,
        maximum=160,
    )
    min_default = min(
        config.pipeline.min_chars_per_line,
        config.pipeline.max_chars_per_line,
    )
    config.pipeline.min_chars_per_line = _wizard_ask_int(
        "Minimum characters per source line",
        default=min_default,
        minimum=10,
        maximum=config.pipeline.max_chars_per_line,
    )
    translation_max_default = (
        config.pipeline.translation_max_chars_per_line
        if config.pipeline.translation_max_chars_per_line is not None
        else 26
    )
    config.pipeline.translation_max_chars_per_line = _wizard_ask_int(
        "Maximum characters per translated line",
        default=translation_max_default,
        minimum=10,
        maximum=160,
    )
    translation_min_default = (
        config.pipeline.translation_min_chars_per_line
        if config.pipeline.translation_min_chars_per_line is not None
        else 16
    )
    translation_min_default = min(
        translation_min_default,
        config.pipeline.translation_max_chars_per_line,
    )
    config.pipeline.translation_min_chars_per_line = _wizard_ask_int(
        "Minimum characters per translated line",
        default=translation_min_default,
        minimum=1,
        maximum=config.pipeline.translation_max_chars_per_line or 160,
    )


def _wizard_step_prompt(config: TransubConfig) -> None:
    if _wizard_ask_bool("Open an editor to modify the translation prompt?", False):
        edited = typer.edit(config.pipeline.prompt_preamble + "\n")
        if edited:
            config.pipeline.prompt_preamble = edited.strip()
        return
    custom = _wizard_ask_text(
        "Custom system prompt (leave blank to keep default)",
        default=config.pipeline.prompt_preamble,
        allow_blank=True,
    )
    if custom and custom.strip():
        config.pipeline.prompt_preamble = custom.strip()


def _prompt_for_config() -> TransubConfig:
    config = TransubConfig()
    steps: List[WizardStep] = [
        WizardStep(
            "whisper-backend",
            "Whisper Backend",
            lambda _cfg: "Choose where Whisper will run.",
            _wizard_step_whisper_backend,
        ),
        WizardStep(
            "whisper-model",
            "Whisper Model",
            lambda cfg: (
                "Suggested models: " + ", ".join(WHISPER_MODEL_SUGGESTIONS.get(cfg.whisper.backend, []))
            )
            if WHISPER_MODEL_SUGGESTIONS.get(cfg.whisper.backend)
            else "Provide the model identifier or path.",
            _wizard_step_whisper_model,
        ),
        WizardStep(
            "whisper-device",
            "Compute Device",
            lambda _cfg: "Optional: cuda / cpu / mps. Leave blank to auto-detect.",
            _wizard_step_whisper_device,
        ),
        WizardStep(
            "whisper-language",
            "Language Hint",
            lambda _cfg: "Enter a language code (for example, en) or leave blank for auto detection.",
            _wizard_step_whisper_language,
        ),
        WizardStep(
            "whisper-tune",
            "Segmentation Tuning",
            lambda _cfg: "Enable this to reduce fragmented subtitles.",
            _wizard_step_whisper_segmentation,
        ),
        WizardStep(
            "whisper-initial",
            "Initial Prompt",
            lambda _cfg: "Optional text to help Whisper with terminology or role-playing.",
            _wizard_step_whisper_initial_prompt,
        ),
        WizardStep(
            "whisper-api",
            "Whisper API",
            lambda _cfg: "Configure your remote transcription endpoint.",
            _wizard_step_whisper_api,
            lambda cfg: cfg.whisper.backend == "api",
        ),
        WizardStep(
            "whisper-cpp",
            "whisper.cpp Options",
            lambda _cfg: "Specify the executable, model path, and extra parameters.",
            _wizard_step_whisper_cpp,
            lambda cfg: cfg.whisper.backend == "cpp",
        ),
        WizardStep(
            "whisper-mlx",
            "mlx-whisper Options",
            lambda _cfg: "Provide optional model directory and advanced arguments.",
            _wizard_step_whisper_mlx,
            lambda cfg: cfg.whisper.backend == "mlx",
        ),
        WizardStep(
            "llm-provider",
            "Translation Model",
            lambda _cfg: "Configure which LLM performs translation.",
            _wizard_step_llm_provider,
        ),
        WizardStep(
            "llm-batch",
            "Translation Parameters",
            lambda _cfg: "Adjust batch size, target language, and style.",
            _wizard_step_llm_batch_and_style,
        ),
        WizardStep(
            "pipeline-output",
            "Output Settings",
            lambda _cfg: "Pick subtitle format, output directory, and caching options.",
            _wizard_step_pipeline_output,
        ),
        WizardStep(
            "pipeline-limits",
            "Subtitle Length Limits",
            lambda _cfg: "Control character counts for source and translated subtitles.",
            _wizard_step_pipeline_limits,
        ),
        WizardStep(
            "prompt",
            "Translation Prompt",
            lambda _cfg: "Edit the system prompt or keep the default template.",
            _wizard_step_prompt,
        ),
    ]

    idx = 0
    history: List[int] = []

    while idx < len(steps):
        step = steps[idx]
        if not step.condition(config):
            idx += 1
            continue

        active_indices = [i for i, s in enumerate(steps) if s.condition(config)]
        step_position = active_indices.index(idx) + 1
        total_steps = len(active_indices)

        console.clear()
        description = step.description(config)
        body = _wizard_step_body(step.heading, description)
        _print_header(subtitle=f"setup {step_position}/{total_steps}", body=body)

        try:
            step.handler(config)
            history.append(idx)
            idx += 1
        except BackAction:
            if history:
                idx = history.pop()
                continue
            console.print("[yellow]You are already at the first step; cannot go back further.[/]")
            _wait_for_enter()
            idx = 0

    return config


def _run_wizard(manager: ConfigManager, allow_overwrite: bool) -> None:
    if manager.exists():
        if not allow_overwrite:
            console.print("Configuration untouched.")
            raise typer.Exit(code=0)
        overwrite = Confirm.ask(
            f"A config already exists at [bold]{manager.path}[/]. Overwrite?", default=False
        )
        if not overwrite:
            console.print("Configuration untouched.")
            raise typer.Exit(code=0)

    config = _prompt_for_config()
    manager.save(config)
    console.print(
        Panel.fit(
            f"Configuration saved to {manager.path}",
            style="green",
        )
    )


def _key_value_table(rows: List[tuple[str, RenderableType]]) -> Table:
    table = Table.grid(padding=(0, 1))
    table.add_column(style="dim", no_wrap=True)
    table.add_column(style="white", ratio=1)
    for key, value in rows:
        table.add_row(key, value)
    return table


def _config_summary_table(config: TransubConfig, path: Path) -> Table:
    whisper = config.whisper
    llm = config.llm
    pipeline = config.pipeline
    rows = [
        ("config", str(path)),
        (
            "whisper",
            (
                f"backend={whisper.backend} | model={whisper.model} | "
                f"device={whisper.device or 'auto'} | language={whisper.language or 'auto'}"
            ),
        ),
        (
            "llm",
            (
                f"provider={llm.provider} | model={llm.model} | target={llm.target_language} | batch={llm.batch_size}"
            ),
        ),
        (
            "pipeline",
            (
                f"format={pipeline.output_format} | audio={pipeline.audio_format} | dir={pipeline.output_dir} | "
                f"keep_audio={_fmt_bool(pipeline.keep_temp_audio)} | save_en={_fmt_bool(pipeline.save_source_subtitles)}"
            ),
        ),
        (
            "limits",
            (
                f"max_line={pipeline.max_chars_per_line} | min_line={pipeline.min_chars_per_line} | "
                f"translated_max={pipeline.translation_max_chars_per_line} | translated_min={pipeline.translation_min_chars_per_line}"
            ),
        ),
    ]
    return _key_value_table(rows)


def _wait_for_enter(message: str = "Press Enter to continue...") -> None:
    console.input(f"[dim]{message}[/]")


def _configure_whisper_backend(config: TransubConfig) -> None:
    whisper = config.whisper
    while True:
        console.clear()
        summary_lines = [
            f"Backend: {whisper.backend}",
            f"Model: {whisper.model}",
            f"Device: {whisper.device or 'auto'} | Language: {whisper.language or 'auto'}",
        ]
        if whisper.backend == "api":
            summary_lines.append(
                f"API url: {whisper.api_url or '(default)'} | key env: {whisper.api_key_env}"
            )
        elif whisper.backend == "cpp":
            summary_lines.append(
                f"cpp_binary: {whisper.cpp_binary} | model_path: {whisper.cpp_model_path or '(none)'} | threads: {whisper.cpp_threads or 'auto'}"
            )
        elif whisper.backend == "mlx":
            summary_lines.append(
                f"mlx_model_dir: {whisper.mlx_model_dir or '(auto)'} | dtype: {whisper.mlx_dtype or 'auto'} | device: {whisper.mlx_device or 'auto'}"
            )
        console.print(
            Panel("\\n".join(summary_lines), title="Whisper backend & model", border_style=THEME_COLOR)
        )

        options = [
            ("Change backend", "backend"),
            ("Change model", "model"),
            ("Change device", "device"),
            ("Change language", "language"),
        ]
        if whisper.backend == "api":
            options.append(("Set API settings", "api"))
        if whisper.backend == "cpp":
            options.append(("Set whisper.cpp options", "cpp"))
        if whisper.backend == "mlx":
            options.append(("Set mlx options", "mlx"))

        console.print("0. Back")
        for idx, (label, _) in enumerate(options, start=1):
            console.print(f"{idx}. {label}")
        choice = Prompt.ask("Select an option", default="0")
        if choice == "0":
            return
        try:
            _, key = options[int(choice) - 1]
        except (ValueError, IndexError):
            _wait_for_enter("Selection out of range. Press Enter to continue...")
            continue

        if key == "backend":
            new_backend = Prompt.ask(
                "Backend",
                choices=["local", "api", "cpp", "mlx"],
                default=whisper.backend,
            )
            if new_backend != whisper.backend:
                whisper.backend = new_backend
                suggestions = WHISPER_MODEL_SUGGESTIONS.get(new_backend, [])
                if suggestions and Confirm.ask(
                    f"Use suggested model '{suggestions[0]}'?",
                    default=True,
                ):
                    whisper.model = suggestions[0]
            continue

        if key == "model":
            suggestions = WHISPER_MODEL_SUGGESTIONS.get(whisper.backend, [])
            if suggestions:
                console.print("Suggested models: " + ", ".join(suggestions))
            whisper.model = Prompt.ask(
                "Model id or path",
                default=whisper.model,
                show_choices=False,
            )
        elif key == "device":
            whisper.device = (
                Prompt.ask(
                    "Preferred device (blank for auto)",
                    default=whisper.device or "",
                    show_choices=False,
                )
                or None
            )
        elif key == "language":
            whisper.language = (
                Prompt.ask(
                    "Source language hint (blank for auto)",
                    default=whisper.language or "",
                    show_choices=False,
                )
                or None
            )
        elif key == "api":
            whisper.api_url = (
                Prompt.ask(
                    "API URL (blank for default)",
                    default=whisper.api_url or "https://api.openai.com/v1/audio/transcriptions",
                    show_choices=False,
                )
                or None
            )
            whisper.api_key_env = Prompt.ask(
                "API key environment variable",
                default=whisper.api_key_env,
            )
        elif key == "cpp":
            whisper.cpp_binary = Prompt.ask(
                "whisper.cpp executable",
                default=whisper.cpp_binary,
                show_choices=False,
            )
            whisper.cpp_model_path = (
                Prompt.ask(
                    "whisper.cpp model path (.bin/.gguf) (blank to clear)",
                    default=whisper.cpp_model_path or "",
                    show_choices=False,
                )
                or None
            )
            threads_raw = Prompt.ask(
                "whisper.cpp threads (blank for auto)",
                default=str(whisper.cpp_threads or ""),
                show_choices=False,
            ).strip()
            whisper.cpp_threads = int(threads_raw) if threads_raw else None
            extra_args_raw = Prompt.ask(
                "Extra whisper.cpp arguments (space separated)",
                default=" ".join(whisper.cpp_extra_args),
                show_choices=False,
            ).strip()
            whisper.cpp_extra_args = shlex.split(extra_args_raw) if extra_args_raw else []
        elif key == "mlx":
            whisper.mlx_model_dir = (
                Prompt.ask(
                    "mlx-whisper model directory (blank for auto)",
                    default=whisper.mlx_model_dir or "",
                    show_choices=False,
                )
                or None
            )
            whisper.mlx_dtype = (
                Prompt.ask(
                    "mlx dtype (auto/float16/float32)",
                    default=whisper.mlx_dtype or "",
                    show_choices=False,
                )
                or None
            )
            whisper.mlx_device = (
                Prompt.ask(
                    "mlx device (auto/mps/cpu)",
                    default=whisper.mlx_device or "",
                    show_choices=False,
                )
                or None
            )
            extra_json = Prompt.ask(
                "Extra mlx-whisper arguments (JSON)",
                default=json.dumps(whisper.mlx_extra_args or {}),
                show_choices=False,
            )
            try:
                whisper.mlx_extra_args = json.loads(extra_json) if extra_json.strip() else {}
            except json.JSONDecodeError:
                _wait_for_enter("Invalid JSON. Press Enter to continue...")


def _configure_whisper_advanced(config: TransubConfig) -> None:
    whisper = config.whisper
    while True:
        console.clear()
        summary_lines = [
            f"Segmentation tuning: {_fmt_bool(whisper.tune_segmentation)} | Temperature: {whisper.temperature if whisper.temperature is not None else 'auto'} | Compression ratio: {whisper.compression_ratio_threshold}",
            f"Logprob threshold: {whisper.logprob_threshold} | No-speech threshold: {whisper.no_speech_threshold}",
            f"Condition on previous text: {_fmt_bool(bool(whisper.condition_on_previous_text))} | Initial prompt: {whisper.initial_prompt or '(none)'}",
        ]
        console.print(
            Panel("\\n".join(summary_lines), title="Whisper advanced parameters", border_style=THEME_COLOR)
        )

        options = [
            ("Toggle segmentation tuning", "seg"),
            ("Set temperature", "temp"),
            ("Set compression ratio threshold", "comp"),
            ("Set logprob threshold", "logprob"),
            ("Set no-speech threshold", "nospeech"),
            ("Set condition on previous text", "condition"),
            ("Edit initial prompt", "prompt"),
        ]
        for idx, (label, _) in enumerate(options, start=1):
            console.print(f"{idx}. {label}")
        console.print("0. Back")
        choice = Prompt.ask("Select an option", default="0")
        if choice == "0":
            return
        try:
            _, key = options[int(choice) - 1]
        except (ValueError, IndexError):
            _wait_for_enter("Selection out of range. Press Enter to continue...")
            continue

        if key == "seg":
            whisper.tune_segmentation = not whisper.tune_segmentation
        elif key == "temp":
            whisper.temperature = _prompt_float_optional(
                "Temperature (blank to keep, 'none' to clear)", whisper.temperature
            )
        elif key == "comp":
            whisper.compression_ratio_threshold = _prompt_float_optional(
                "Compression ratio threshold (blank to keep)",
                whisper.compression_ratio_threshold,
            )
        elif key == "logprob":
            whisper.logprob_threshold = _prompt_float_optional(
                "Logprob threshold (blank to keep)", whisper.logprob_threshold
            )
        elif key == "nospeech":
            whisper.no_speech_threshold = _prompt_float_optional(
                "No-speech threshold (blank to keep)", whisper.no_speech_threshold
            )
        elif key == "condition":
            whisper.condition_on_previous_text = Confirm.ask(
                "Condition on previous text?",
                default=bool(whisper.condition_on_previous_text)
                if whisper.condition_on_previous_text is not None
                else True,
            )
        elif key == "prompt":
            whisper.initial_prompt = (
                Prompt.ask(
                    "Initial prompt (blank to clear)",
                    default=whisper.initial_prompt or "",
                    show_choices=False,
                )
                or None
            )


def _configure_llm(config: TransubConfig) -> None:
    llm = config.llm
    while True:
        console.clear()
        summary_lines = [
            f"Provider: {llm.provider} | Model: {llm.model}",
            f"Target: {llm.target_language} | Batch size: {llm.batch_size}",
            f"Temperature: {llm.temperature} | Max retries: {llm.max_retries} | Timeout: {llm.request_timeout}s",
            f"API base: {llm.api_base or '(default)'} | API key env: {llm.api_key_env} | Style: {llm.style or '(none)'}",
        ]
        console.print(Panel("\\n".join(summary_lines), title="Translation LLM", border_style=THEME_COLOR))

        options = [
            ("Change provider", "provider"),
            ("Change model", "model"),
            ("Set API base", "base"),
            ("Set API key env", "key"),
            ("Set batch size", "batch"),
            ("Set temperature", "temp"),
            ("Set max retries", "retries"),
            ("Set request timeout", "timeout"),
            ("Set target language", "target"),
            ("Set style descriptor", "style"),
        ]
        for idx, (label, _) in enumerate(options, start=1):
            console.print(f"{idx}. {label}")
        console.print("0. Back")
        choice = Prompt.ask("Select an option", default="0")
        if choice == "0":
            return
        try:
            _, key = options[int(choice) - 1]
        except (ValueError, IndexError):
            _wait_for_enter("Selection out of range. Press Enter to continue...")
            continue

        if key == "provider":
            llm.provider = Prompt.ask("Provider name", default=llm.provider)
        elif key == "model":
            llm.model = Prompt.ask("Model id", default=llm.model)
        elif key == "base":
            llm.api_base = (
                Prompt.ask(
                    "API base URL (blank for default)",
                    default=llm.api_base or "",
                    show_choices=False,
                )
                or None
            )
        elif key == "key":
            llm.api_key_env = Prompt.ask("API key environment variable", default=llm.api_key_env)
        elif key == "batch":
            llm.batch_size = _prompt_int(
                "Lines per translation batch",
                default=llm.batch_size,
                minimum=1,
                maximum=50,
            )
        elif key == "temp":
            llm.temperature = _prompt_float("Temperature", llm.temperature)
        elif key == "retries":
            llm.max_retries = _prompt_int(
                "Max retries",
                default=llm.max_retries,
                minimum=0,
                maximum=10,
            )
        elif key == "timeout":
            llm.request_timeout = _prompt_float(
                "Request timeout (seconds)",
                llm.request_timeout,
            )
        elif key == "target":
            llm.target_language = Prompt.ask(
                "Target language code",
                default=llm.target_language,
            )
        elif key == "style":
            llm.style = (
                Prompt.ask(
                    "Style descriptor (blank to clear)",
                    default=llm.style or "",
                    show_choices=False,
                )
                or None
            )


def _configure_pipeline(config: TransubConfig) -> None:
    pipeline = config.pipeline
    while True:
        console.clear()
        summary_lines = [
            f"Format: {pipeline.output_format} | Audio: {pipeline.audio_format}",
            f"Output dir: {pipeline.output_dir}",
            f"Keep temp audio: {_fmt_bool(pipeline.keep_temp_audio)} | Save source subtitles: {_fmt_bool(pipeline.save_source_subtitles)}",
            f"Max line chars: {pipeline.max_chars_per_line} (min {pipeline.min_chars_per_line})",
            f"Translated max: {pipeline.translation_max_chars_per_line} (min {pipeline.translation_min_chars_per_line})",
            f"Timing trim: {pipeline.timing_trim_seconds}s | Offset: {pipeline.timing_offset_seconds}s | Minimum duration: {pipeline.min_line_duration}s",
            f"Trim punctuation: {_fmt_bool(pipeline.remove_trailing_punctuation)} | CJK spacing: {_fmt_bool(pipeline.normalize_cjk_spacing)}",
            f"Refine source export: {_fmt_bool(pipeline.refine_source_subtitles)}",
        ]
        console.print(Panel("\\n".join(summary_lines), title="Pipeline & output", border_style=THEME_COLOR))

        options = [
            ("Set output format", "format"),
            ("Set audio format", "audio"),
            ("Set output directory", "dir"),
            ("Toggle keep temp audio", "keep"),
            ("Toggle save source subtitles", "save"),
            ("Set max characters per line", "max_line"),
            ("Set min characters per line", "min_line"),
            ("Set translated max characters", "tmax"),
            ("Set translated min characters", "tmin"),
            ("Set timing trim seconds", "trim"),
            ("Set timing offset seconds", "offset"),
            ("Set minimum line duration", "duration"),
            ("Toggle remove trailing punctuation", "punct"),
            ("Toggle CJK-Latin spacing", "spacing"),
            ("Toggle refine source export", "refine_en"),
            ("Edit translation system prompt", "prompt"),
        ]
        for idx, (label, _) in enumerate(options, start=1):
            console.print(f"{idx}. {label}")
        console.print("0. Back")
        choice = Prompt.ask("Select an option", default="0")
        if choice == "0":
            return
        try:
            _, key = options[int(choice) - 1]
        except (ValueError, IndexError):
            _wait_for_enter("Selection out of range. Press Enter to continue...")
            continue

        if key == "format":
            pipeline.output_format = Prompt.ask(
                "Output format",
                choices=["srt", "vtt"],
                default=pipeline.output_format,
            )
        elif key == "audio":
            pipeline.audio_format = Prompt.ask(
                "Intermediate audio format",
                choices=["wav", "mp3", "flac", "m4a", "ogg"],
                default=pipeline.audio_format,
            )
        elif key == "dir":
            pipeline.output_dir = Prompt.ask(
                "Output directory",
                default=pipeline.output_dir,
                show_choices=False,
            )
        elif key == "keep":
            pipeline.keep_temp_audio = Confirm.ask(
                "Keep intermediate audio file?",
                default=pipeline.keep_temp_audio,
            )
        elif key == "save":
            pipeline.save_source_subtitles = Confirm.ask(
                "Save source subtitles?",
                default=pipeline.save_source_subtitles,
            )
        elif key == "max_line":
            pipeline.max_chars_per_line = _prompt_int(
                "Max characters per subtitle line",
                default=pipeline.max_chars_per_line,
                minimum=20,
                maximum=160,
            )
            if pipeline.min_chars_per_line > pipeline.max_chars_per_line:
                pipeline.min_chars_per_line = pipeline.max_chars_per_line
        elif key == "min_line":
            pipeline.min_chars_per_line = _prompt_int(
                "Min characters per subtitle line",
                default=pipeline.min_chars_per_line,
                minimum=10,
                maximum=pipeline.max_chars_per_line,
            )
        elif key == "tmax":
            pipeline.translation_max_chars_per_line = _prompt_int(
                "Max characters per translated line",
                default=pipeline.translation_max_chars_per_line or 36,
                minimum=10,
                maximum=160,
            )
            if (
                pipeline.translation_min_chars_per_line
                and pipeline.translation_min_chars_per_line > pipeline.translation_max_chars_per_line
            ):
                pipeline.translation_min_chars_per_line = pipeline.translation_max_chars_per_line
        elif key == "tmin":
            max_chars = pipeline.translation_max_chars_per_line or 160
            pipeline.translation_min_chars_per_line = _prompt_int(
                "Min characters per translated line",
                default=pipeline.translation_min_chars_per_line or 4,
                minimum=1,
                maximum=max_chars,
            )
        elif key == "trim":
            pipeline.timing_trim_seconds = _prompt_float(
                "Timing trim seconds",
                pipeline.timing_trim_seconds,
            )
        elif key == "offset":
            pipeline.timing_offset_seconds = _prompt_float(
                "Timing offset seconds (positive delays subtitles)",
                pipeline.timing_offset_seconds,
            )
        elif key == "duration":
            pipeline.min_line_duration = _prompt_float(
                "Minimum line duration (seconds)",
                pipeline.min_line_duration,
            )
        elif key == "punct":
            pipeline.remove_trailing_punctuation = Confirm.ask(
                "Remove trailing punctuation?",
                default=pipeline.remove_trailing_punctuation,
            )
        elif key == "spacing":
            pipeline.normalize_cjk_spacing = Confirm.ask(
                "Insert spaces between CJK characters and Latin/digit text?",
                default=pipeline.normalize_cjk_spacing,
            )
        elif key == "refine_en":
            pipeline.refine_source_subtitles = Confirm.ask(
                "Re-refine source subtitles when exporting?",
                default=pipeline.refine_source_subtitles,
            )
        elif key == "prompt":
            edited = typer.edit(pipeline.prompt_preamble + "\n")
            pipeline.prompt_preamble = (edited or pipeline.prompt_preamble).strip()

def _fmt_bool(value: bool, true_label: str = "yes", false_label: str = "no") -> str:
    return true_label if value else false_label


def _prompt_float(question: str, default: float) -> float:
    while True:
        raw = Prompt.ask(question, default=str(default), show_choices=False)
        try:
            return float(raw)
        except ValueError:
            console.print("[red]Please enter a numeric value.[/]")


def _prompt_float_optional(question: str, current: Optional[float]) -> Optional[float]:
    default = "" if current is None else str(current)
    while True:
        raw = Prompt.ask(question, default=default, show_choices=False).strip()
        if not raw:
            return current
        if raw.lower() in {"none", "null"}:
            return None
        try:
            return float(raw)
        except ValueError:
            console.print("[red]Please enter a numeric value or 'none'.[/]")


def _prompt_int(question: str, default: int, minimum: int, maximum: int) -> int:
    while True:
        raw = Prompt.ask(question, default=str(default))
        try:
            value = int(raw)
        except ValueError:
            console.print("[red]Please enter an integer value.[/]")
            continue
        if value < minimum or value > maximum:
            console.print(
                f"[red]Value must be between {minimum} and {maximum}.[/]"
            )
            continue
        return value


def _prompt_optional_int(
    question: str,
    minimum: int,
) -> Optional[int]:
    raw = Prompt.ask(question, default="")
    if not raw.strip():
        return None
    try:
        value = int(raw)
    except ValueError:
        console.print("[red]Please enter an integer or leave blank.[/]")
        return _prompt_optional_int(question, minimum)
    if value < minimum:
        console.print(f"[red]Value must be at least {minimum}.[/]")
        return _prompt_optional_int(question, minimum)
    return value


def _prompt_non_empty(question: str, default: str) -> str:
    while True:
        raw = Prompt.ask(question, default=default)
        if raw.strip():
            return raw
        console.print("[red]This field cannot be empty. Please try again.[/]")


def _prompt_json_dict(question: str, default: str) -> dict:
    while True:
        raw = Prompt.ask(question, default=default)
        if not raw.strip():
            return {}
        try:
            value = json.loads(raw)
        except json.JSONDecodeError:
            console.print("[red]Please enter a valid JSON object.[/]")
            continue
        if not isinstance(value, dict):
            console.print("[red]A JSON object (key/value pairs) is required.[/]")
            continue
        return value


def _write_document(
    document: SubtitleDocument,
    target_dir: Path,
    stem: str,
    suffix: str,
    output_format: str,
) -> Path:
    target_dir.mkdir(parents=True, exist_ok=True)
    filename = f"{stem}{suffix}.{output_format}"
    output_path = target_dir / filename
    if output_format == "srt":
        output_path.write_text(document.to_srt(), encoding="utf-8")
    elif output_format == "vtt":
        output_path.write_text(document.to_vtt(), encoding="utf-8")
    else:  # pragma: no cover - guarded by config validation
        raise ValueError(f"Unsupported output format: {output_format}")
    return output_path


def _language_suffix(language: Optional[str]) -> str:
    if not language:
        return ""
    normalized = language.strip().lower().replace("-", "_").replace(" ", "_")
    normalized = normalized.strip("._")
    if not normalized:
        return ""
    return f".{normalized}"


def _source_language_suffix(language: Optional[str]) -> str:
    if not language or language.strip().lower() in {"", "auto"}:
        return ".src"
    suffix = _language_suffix(language)
    return suffix or ".src"


def _cleanup_audio_file(work_dir: Path, audio_path: Optional[Path]) -> None:
    if not audio_path:
        return
    try:
        audio_path.unlink(missing_ok=True)
        if not any(work_dir.iterdir()):
            work_dir.rmdir()
    except OSError:
        pass


def _offer_failure_cleanup(
    work_dir: Path,
    state: PipelineState,
    audio_path: Optional[Path],
    translations_path: Optional[Path],
    interrupted: bool,
) -> None:
    cached_paths = []
    current_segments = state.get_segments_path()
    if audio_path and audio_path.exists():
        cached_paths.append(audio_path)
    if current_segments and current_segments.exists():
        cached_paths.append(current_segments)
    if translations_path and translations_path.exists():
        cached_paths.append(translations_path)
    if state.path.exists():
        cached_paths.append(state.path)

    if not cached_paths:
        return

    prompt_text = "Pipeline did not finish. Clear cached data before retrying?"
    if interrupted:
        prompt_text = "Interrupted detected. Clear cached data now?"
    try:
        should_cleanup = Confirm.ask(prompt_text, default=False)
    except Exception:
        return

    if not should_cleanup:
        console.print("[yellow]Cached data kept. Resume the run after fixing the issue.[/]")
        return

    if audio_path and audio_path.exists():
        _cleanup_audio_file(work_dir, audio_path)
    if current_segments and current_segments.exists():
        try:
            current_segments.unlink()
        except OSError:
            pass
    if translations_path and translations_path.exists():
        try:
            translations_path.unlink()
        except OSError:
            pass
    state.clear()
    try:
        if work_dir.exists() and not any(work_dir.iterdir()):
            work_dir.rmdir()
    except OSError:
        pass
    console.print("[green]Cached data cleared.[/]")


if __name__ == "__main__":
    app()



from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, Optional
from textwrap import dedent

try:  # Python 3.11+
    import tomllib as tomli  # type: ignore[import-not-found]
except ModuleNotFoundError:  # pragma: no cover - fallback for <3.11
    import tomli  # type: ignore[no-redef]

import tomli_w
from pydantic import BaseModel, Field, ValidationError, model_validator

CONFIG_FILENAME = "transub.conf"

DEFAULT_TRANSLATION_PROMPT = dedent(
    """\
    # Role: Senior Subtitle Translator
    You are an experienced subtitle translator who delivers clear, natural ${targetLanguage} subtitles.

    # Guidelines
    1. Keep each subtitle line independent; never merge or split entries.
    2. Prefer conversational ${targetLanguage} that reads well in subtitles.
    3. Use punctuation (comma, period, ellipsis) to preserve rhythm and tone.
    4. Translate terminology accurately and keep it consistent across lines.
    5. Use straight quotes instead of curly quotes.
    6. When the translation mixes different scripts (for example, Latin letters within Chinese text), add spacing where appropriate so the subtitle stays legible.

    # Output Format
    1. Return a JSON object that mirrors the input keys (IDs) and only translates the values.
    2. Do not add commentary or text outside the JSON object.
    3. Ensure the JSON is syntactically valid and contains the same number of entries as the input.

    Review your work to confirm the translation is fluent, faithful to the source, and grammatically correct. Adapt passive/active voice as needed for natural ${targetLanguage} subtitles.

    # Example

    Input:
    {"0": "Welcome to China", "1": "China is a beautiful country"}

    Output:
    {"0": "欢迎来到中国", "1": "中国是一个美丽的国家"}
    """
).strip()


class WhisperConfig(BaseModel):
    """Configuration for Whisper transcription."""

    backend: str = Field(
        default="local", description="Which backend to use: local, api, cpp, or mlx"
    )
    execution_mode: str = Field(
        default="internal",
        description="Execution strategy: 'internal' (python import) or 'external' (cli subprocess)",
    )
    cli_path: str | None = Field(
        default=None,
        description="Path to the CLI executable for 'external' mode",
    )
    model: str = Field(default="base", description="Whisper model size")
    device: str | None = Field(
        default=None,
        description="Override compute device, e.g. cuda, cpu, mps",
    )
    api_url: str | None = Field(
        default=None,
        description="Custom transcription API endpoint when backend=api",
    )
    api_key_env: str = Field(
        default="OPENAI_API_KEY",
        description="Environment variable storing the speech-to-text API key",
    )
    cpp_binary: str = Field(
        default="whisper-cpp",
        description="Executable name or path for the whisper.cpp CLI",
    )
    cpp_model_path: str | None = Field(
        default=None,
        description="Path to the ggml/gguf model file when using whisper.cpp backend",
    )
    cpp_threads: int | None = Field(
        default=None,
        description="Optional number of threads for whisper.cpp",
        ge=1,
    )
    cpp_extra_args: list[str] = Field(
        default_factory=list,
        description="Additional CLI arguments for whisper.cpp backend",
    )
    mlx_model_dir: str | None = Field(
        default=None,
        description="Directory containing mlx-whisper converted model weights",
    )
    mlx_dtype: str | None = Field(
        default=None,
        description="Computation dtype for mlx-whisper (auto, float16, float32, etc.)",
    )
    mlx_device: str | None = Field(
        default=None,
        description="Target device for mlx-whisper (cpu, mps). Defaults to mlx auto detect.",
    )
    mlx_extra_args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Additional keyword arguments forwarded to mlx-whisper",
    )
    language: str | None = Field(
        default="en",
        description="Language hint passed to Whisper when transcribing",
    )
    extra_args: Dict[str, Any] = Field(
        default_factory=dict,
        description="Additional backend-specific arguments",
    )
    tune_segmentation: bool = Field(
        default=True,
        description="Apply recommended Whisper segmentation parameters to reduce fragmenting.",
    )
    temperature: float | None = Field(
        default=0.0,
        ge=0.0,
        le=1.0,
        description="Sampling temperature for local Whisper decoding (None to keep library default).",
    )
    compression_ratio_threshold: float | None = Field(
        default=2.6,
        ge=0.0,
        description="Compression ratio threshold before a segment is discarded (None to keep default).",
    )
    logprob_threshold: float | None = Field(
        default=-1.0,
        description="Minimum average log probability for valid decoding (None to keep default).",
    )
    no_speech_threshold: float | None = Field(
        default=0.3,
        ge=0.0,
        le=1.0,
        description="Probability threshold for classifying a segment as silence (None to keep default).",
    )
    condition_on_previous_text: bool | None = Field(
        default=True,
        description="Condition decoding on previous text to maintain context.",
    )
    initial_prompt: str | None = Field(
        default=None,
        description="Optional initial prompt to guide Whisper's segmentation and terminology.",
    )

    @model_validator(mode="after")
    def validate_backend(self) -> "WhisperConfig":
        backend = self.backend.lower()
        if backend not in {"local", "api", "cpp", "mlx"}:
            raise ValueError("backend must be 'local', 'api', 'cpp', or 'mlx'")
        object.__setattr__(self, "backend", backend)
        if backend == "cpp" and not self.cpp_model_path:
            raise ValueError("cpp_model_path must be set when backend is 'cpp'")
        return self


class LLMConfig(BaseModel):
    """Configuration for the translation LLM."""

    provider: str = Field(default="openai", description="LLM provider identifier")
    model: str = Field(default="gpt-4o-mini", description="Model name")
    api_base: str | None = Field(
        default=None, description="Custom endpoint base URL if needed"
    )
    api_key_env: str = Field(
        default="OPENAI_API_KEY",
        description="Environment variable storing the API key",
    )
    temperature: float = Field(default=0.2, ge=0, le=2)
    max_retries: int = Field(default=3, ge=0, le=10)
    request_timeout: float = Field(default=60.0, gt=0)
    batch_size: int = Field(
        default=5,
        description="Number of subtitle lines per translation request",
        ge=1,
        le=50,
    )
    target_language: str = Field(default="zh", description="Target translation language")
    style: str | None = Field(
        default="Simplified Chinese",
        description="Optional description of translation tone/style",
    )


class PipelineConfig(BaseModel):
    """Configuration for pipeline level options."""

    output_format: str = Field(
        default="srt",
        description="Subtitle output format (srt or vtt)",
    )
    audio_format: str = Field(
        default="wav",
        description="Intermediate audio format for whisper input",
    )
    keep_temp_audio: bool = Field(
        default=False, description="Keep intermediate extracted audio file"
    )
    output_dir: str | None = Field(
        default=None,
        description="Directory for generated subtitle files",
    )
    save_source_subtitles: bool = Field(
        default=True,
        description="Whether to save the intermediate source-language subtitles",
    )
    max_chars_per_line: int = Field(
        default=60,
        ge=20,
        le=160,
        description="Maximum characters per subtitle line after refinement",
    )
    min_chars_per_line: int = Field(
        default=25,
        ge=10,
        le=120,
        description="Preferred minimum characters per subtitle line",
    )
    timing_trim_seconds: float = Field(
        default=0.15,
        ge=0.0,
        le=1.0,
        description="Seconds trimmed from start/end of each subtitle for tighter timing",
    )
    timing_offset_seconds: float = Field(
        default=0.0,
        ge=-1.0,
        le=1.0,
        description="Shift applied to subtitle start/end times (positive delays subtitles, negative moves earlier)",
    )
    min_line_duration: float = Field(
        default=0.6,
        ge=0.1,
        le=5.0,
        description="Minimum duration (seconds) each subtitle should stay on screen",
    )
    remove_trailing_punctuation: bool = Field(
        default=True,
        description="Remove trailing punctuation from translated subtitles",
    )
    normalize_cjk_spacing: bool = Field(
        default=True,
        description="Insert spaces between CJK characters and Latin/digit sequences for readability",
    )
    prompt_preamble: str = Field(
        default=DEFAULT_TRANSLATION_PROMPT,
        description="System prompt prepended to LLM translation requests",
    )
    translation_max_chars_per_line: int | None = Field(
        default=26,
        ge=10,
        le=160,
        description="Maximum characters per translated subtitle line (None to disable post-translation reflow)",
    )
    translation_min_chars_per_line: int | None = Field(
        default=16,
        ge=1,
        le=120,
        description="Preferred minimum characters per translated subtitle line",
    )
    refine_source_subtitles: bool = Field(
        default=False,
        description="Whether to re-refine and reflow source subtitles on export",
    )

    @model_validator(mode="after")
    def validate_format(self) -> "PipelineConfig":
        fmt = self.output_format.lower()
        if fmt not in {"srt", "vtt"}:
            raise ValueError("output_format must be 'srt' or 'vtt'")
        object.__setattr__(self, "output_format", fmt)
        audio_fmt = self.audio_format.lower()
        if audio_fmt not in {"wav", "mp3", "flac", "m4a", "ogg"}:
            raise ValueError(
                "audio_format must be one of wav, mp3, flac, m4a, ogg"
            )
        object.__setattr__(self, "audio_format", audio_fmt)
        if self.min_chars_per_line > self.max_chars_per_line:
            raise ValueError("min_chars_per_line cannot exceed max_chars_per_line")
        if (
            self.translation_max_chars_per_line is not None
            and self.translation_min_chars_per_line is not None
            and self.translation_min_chars_per_line > self.translation_max_chars_per_line
        ):
            raise ValueError(
                "translation_min_chars_per_line cannot exceed translation_max_chars_per_line"
            )
        return self


class TransubConfig(BaseModel):
    """Top-level configuration model for Transub."""

    whisper: WhisperConfig = Field(default_factory=WhisperConfig)
    llm: LLMConfig = Field(default_factory=LLMConfig)
    pipeline: PipelineConfig = Field(default_factory=PipelineConfig)


@dataclass
class ConfigManager:
    """Handles loading and saving the CLI configuration."""

    path: Path

    def exists(self) -> bool:
        return self.path.exists()

    def load(self) -> TransubConfig:
        if not self.exists():
            raise FileNotFoundError(
                f"Configuration file not found: {self.path}. Run 'transub init' first."
            )
        with self.path.open("rb") as fh:
            raw = tomli.load(fh)
        try:
            return TransubConfig.model_validate(raw)
        except ValidationError as exc:
            raise ValueError(
                "Configuration file is invalid. Please fix the errors or re-run 'transub init'."
            ) from exc

    def save(self, config: TransubConfig) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open("wb") as fh:
            tomli_w.dump(
                config.model_dump(mode="json", exclude_none=True), fh
            )

    @classmethod
    def default_path(cls) -> Path:
        # Environment variable overrides everything so packaged installs can
        # point at a shared config location.
        env_path = os.getenv("TRANSUB_CONFIG")
        if env_path:
            return Path(env_path).expanduser()

        return Path.home() / ".transub" / CONFIG_FILENAME


def load_or_create_default(path: Optional[Path] = None) -> TransubConfig:
    """Load config if present, otherwise write and return defaults."""

    cfg_path = path or ConfigManager.default_path()
    manager = ConfigManager(cfg_path)
    if manager.exists():
        return manager.load()
    default_config = TransubConfig()
    manager.save(default_config)
    return default_config


__all__ = [
    "CONFIG_FILENAME",
    "DEFAULT_TRANSLATION_PROMPT",
    "ConfigManager",
    "TransubConfig",
    "WhisperConfig",
    "LLMConfig",
    "PipelineConfig",
    "load_or_create_default",
]



from __future__ import annotations

from dataclasses import dataclass
from datetime import timedelta
from typing import Iterable, List
import re

_SOFT_PUNCT_PATTERN = re.compile(r"[，,；;、]")
_HARD_PUNCT_PATTERN = re.compile(r"[。\.!?！？]")
_SOFT_BREAK_CHARS = {"，", ",", "；", ";", "、"}
_HARD_BREAK_CHARS = {"。", ".", "?", "？", "!", "！"}
_SHORT_DANGLERS = {"a", "an", "the", "and", "or", "but", "so", "to", "for", "of", "in", "on", "at", "I", "i"}
_CHINESE_CONNECTIVE_SUFFIXES = ("但是", "不过", "然而", "可是", "所以", "而且")
_CHINESE_CONNECTIVE_CHARS = {"和", "或", "但", "而", "且", "并", "却", "又"}
_ENGLISH_CONNECTIVES = {"and", "or", "but", "so", "yet", "though"}
_TRAILING_PUNCTUATION = "。！？?!?,.;；:：…，、"
_CLOSING_WRAPPERS = "\"'”’）)】》〉］]」』＞｝】"
_CJK_LATIN_LEFT_PATTERN = re.compile(r"([\u4e00-\u9fff])([A-Za-z0-9])")
_CJK_LATIN_RIGHT_PATTERN = re.compile(r"([A-Za-z0-9])([\u4e00-\u9fff])")


def _split_text_for_limits(text: str, max_chars: int, min_chars: int) -> List[str]:
    """Split text into chunks that respect max length while avoiding tiny tails."""

    remaining = text.strip()
    parts: List[str] = []
    while remaining:
        if len(remaining) <= max_chars:
            parts.append(remaining.strip())
            break
        window = remaining[: max_chars]
        split = None
        hard_candidates = [match.end() for match in _HARD_PUNCT_PATTERN.finditer(window)]
        if hard_candidates:
            candidate = hard_candidates[-1]
            if candidate >= min_chars:
                split = candidate
        if split is None:
            soft_candidates = [match.end() for match in _SOFT_PUNCT_PATTERN.finditer(window)]
            if soft_candidates:
                candidate = soft_candidates[-1]
                if candidate >= min_chars:
                    split = candidate
        if split is None:
            space_idx = window.rfind(" ")
            if space_idx >= max(min_chars, 8):
                split = space_idx + 1

        if split is None or split <= 0:
            split = max_chars

        head = remaining[:split].rstrip()
        tail = remaining[split:].lstrip()

        if not head:
            head = remaining[:max_chars].rstrip()
            tail = remaining[max_chars:].lstrip()

        # Avoid leaving a dangling short word at the start of the tail.
        if tail:
            first_word = tail.split(" ", 1)[0]
            if first_word.lower() in _SHORT_DANGLERS and len(head) > min_chars:
                head = f"{head} {first_word}".rstrip()
                tail = tail[len(first_word) :].lstrip()

        if tail:
            original_head = head

            # Move trailing Chinese connector suffixes to the next chunk.
            for suffix in _CHINESE_CONNECTIVE_SUFFIXES:
                if head.endswith(suffix) and len(head) > len(suffix):
                    head = head[: -len(suffix)].rstrip()
                    tail = f"{suffix}{tail}".lstrip()
                    break

            # Single-character Chinese connectors.
            if head and head[-1] in _CHINESE_CONNECTIVE_CHARS:
                connector = head[-1]
                head = head[:-1].rstrip()
                tail = f"{connector}{tail}".lstrip()

            # English connector words.
            if head:
                tokens = head.split()
                if tokens and tokens[-1].lower() in _ENGLISH_CONNECTIVES:
                    connector = tokens.pop()
                    head = " ".join(tokens).rstrip()
                    tail = f"{connector} {tail}".strip()

            if not head:
                head = original_head

        parts.append(head)
        remaining = tail

    # Merge trailing short fragment with the previous chunk.
    if len(parts) >= 2 and len(parts[-1]) < min_chars:
        parts[-2] = (parts[-2].rstrip() + " " + parts[-1]).strip()
        parts.pop()

    return [part.strip() for part in parts if part.strip()]


def _allocate_timings(start: float, end: float, count: int) -> List[tuple[float, float]]:
    """Evenly split timing between child segments (adjust later by adjust_timing)."""

    if count <= 0:
        return []
    if count == 1 or end <= start:
        return [(start, end)] * count
    duration = end - start
    slice_seconds = duration / count
    segments: List[tuple[float, float]] = []
    current = start
    for idx in range(count):
        if idx == count - 1:
            segment_end = end
        else:
            segment_end = current + slice_seconds
        segments.append((current, max(segment_end, current)))
        current = segment_end
    return segments


def _combine_text(left: str, right: str) -> str:
    left = left.rstrip()
    right = right.lstrip()
    if not left:
        return right
    if not right:
        return left

    left_last = left[-1]
    right_first = right[0]

    if left_last.isascii() and right_first.isascii():
        if left_last.isalnum() and right_first.isalnum():
            return f"{left} {right}"
        if left_last in ".?!," and right_first.isalnum():
            return f"{left} {right}"
    return f"{left}{right}"


@dataclass
class SubtitleLine:
    """Represents a single subtitle line with timing."""

    index: int
    start: float  # seconds
    end: float  # seconds
    text: str

    def to_srt_block(self) -> str:
        start_ts = format_timestamp(self.start)
        end_ts = format_timestamp(self.end)
        return f"{self.index}\n{start_ts} --> {end_ts}\n{self.text}\n"

    def to_vtt_block(self) -> str:
        start_ts = format_timestamp(self.start, separator=".")
        end_ts = format_timestamp(self.end, separator=".")
        return f"{start_ts} --> {end_ts}\n{self.text}\n"


@dataclass
class SubtitleDocument:
    """Collection of subtitle lines."""

    lines: List[SubtitleLine]

    def to_srt(self) -> str:
        blocks = [line.to_srt_block().strip("\n") for line in self.lines]
        if not blocks:
            return ""
        return "\n\n".join(blocks) + "\n"

    def to_vtt(self) -> str:
        header = "WEBVTT\n\n"
        blocks = [line.to_vtt_block().strip("\n") for line in self.lines]
        body = "\n\n".join(blocks) if blocks else ""
        return header + body + "\n"

    def chunk(self, size: int) -> Iterable[List[SubtitleLine]]:
        chunk: List[SubtitleLine] = []
        for line in self.lines:
            chunk.append(line)
            if len(chunk) >= size:
                yield chunk
                chunk = []
        if chunk:
            yield chunk

    @classmethod
    def from_whisper_segments(cls, segments: Iterable[dict]) -> "SubtitleDocument":
        lines = []
        for idx, segment in enumerate(segments, start=1):
            text = (segment.get("text") or "").strip()
            start = float(segment.get("start"))
            end = float(segment.get("end"))
            lines.append(SubtitleLine(index=idx, start=start, end=end, text=text))
        return cls(lines=lines)

    def to_serializable(self) -> List[dict]:
        return [
            {
                "index": line.index,
                "start": line.start,
                "end": line.end,
                "text": line.text,
            }
            for line in self.lines
        ]

    @classmethod
    def from_serialized(cls, data: Iterable[dict]) -> "SubtitleDocument":
        lines = []
        for idx, item in enumerate(data, start=1):
            lines.append(
                SubtitleLine(
                    index=int(item.get("index", idx)),
                    start=float(item.get("start", 0.0)),
                    end=float(item.get("end", 0.0)),
                    text=str(item.get("text", "")).strip(),
                )
            )
        return cls(lines=lines)

    @classmethod
    def from_srt(cls, content: str) -> "SubtitleDocument":
        blocks = re.split(r"\n\s*\n", content.strip())
        lines: List[SubtitleLine] = []
        for block in blocks:
            parts = [part.strip("\ufeff") for part in block.strip().splitlines()]
            if len(parts) < 3:
                continue
            try:
                index = int(parts[0].strip())
            except ValueError:
                continue
            timing = parts[1]
            if "-->" not in timing:
                continue
            start_raw, end_raw = [item.strip() for item in timing.split("-->", 1)]
            try:
                start = parse_timestamp(start_raw)
                end = parse_timestamp(end_raw)
            except ValueError:
                continue
            text = "\n".join(parts[2:]).strip()
            lines.append(SubtitleLine(index=index, start=start, end=end, text=text))
        return cls(lines=lines)

    def refine(self, max_chars: int = 60, min_chars: int = 25) -> "SubtitleDocument":
        if not self.lines:
            return SubtitleDocument(lines=[])

        split_lines: List[SubtitleLine] = []
        for line in self.lines:
            chunks = _split_text_for_limits(line.text, max_chars, min_chars)
            timings = _allocate_timings(line.start, line.end, len(chunks))
            for chunk_text, (chunk_start, chunk_end) in zip(chunks, timings, strict=True):
                split_lines.append(
                    SubtitleLine(
                        index=0,
                        start=chunk_start,
                        end=chunk_end,
                        text=chunk_text,
                    )
                )

        # Merge neighboring lines when chunks are too short or end on soft punctuation.
        merged: List[SubtitleLine] = []
        for line in split_lines:
            if not merged:
                merged.append(line)
                continue

            previous = merged[-1]
            combined_text = _combine_text(previous.text, line.text)
            prev_trimmed = previous.text.rstrip()
            prev_ends_soft = prev_trimmed.endswith(tuple(_SOFT_BREAK_CHARS))
            prev_ends_hard = prev_trimmed.endswith(tuple(_HARD_BREAK_CHARS))
            prev_last_token = prev_trimmed.split()[-1].lower() if prev_trimmed.split() else ""
            prev_ends_connector = (
                any(prev_trimmed.endswith(sfx) for sfx in _CHINESE_CONNECTIVE_SUFFIXES)
                or (prev_trimmed and prev_trimmed[-1] in _CHINESE_CONNECTIVE_CHARS)
                or prev_last_token in _ENGLISH_CONNECTIVES
            )
            needs_merge_for_length = len(line.text) < min_chars and len(combined_text) <= max_chars
            needs_merge_for_soft_break = (
                prev_ends_soft
                and not prev_ends_hard
                and len(combined_text) <= max_chars
            )
            needs_merge_for_connector = prev_ends_connector and len(combined_text) <= max_chars

            if needs_merge_for_length or needs_merge_for_soft_break or needs_merge_for_connector:
                merged[-1] = SubtitleLine(
                    index=0,
                    start=previous.start,
                    end=line.end,
                    text=combined_text,
                )
            else:
                merged.append(line)

        # Reindex sequentially to keep downstream translation maps stable.
        reindexed = [
            SubtitleLine(index=idx, start=line.start, end=line.end, text=line.text)
            for idx, line in enumerate(merged, start=1)
        ]
        return SubtitleDocument(lines=reindexed)

    def adjust_timing(
        self,
        trim: float,
        min_duration: float = 0.6,
        gap: float = 0.04,
        offset: float = 0.0,
    ) -> "SubtitleDocument":
        if trim <= 0 and offset == 0:
            return self
        adjusted: List[SubtitleLine] = []
        lines = self.lines
        prev_end_base: float | None = None

        for idx, line in enumerate(lines):
            start = line.start
            end = line.end
            original_duration = max(line.end - line.start, 0.0)
            if original_duration <= 0:
                original_duration = min_duration

            # Respect previous line gap.
            if prev_end_base is not None and start < prev_end_base + gap:
                start = prev_end_base + gap

            # Only trim leading edge if there is slack beyond the required gap.
            if trim > 0 and prev_end_base is not None:
                slack_before = max(0.0, start - max(prev_end_base + gap, line.start))
                start = min(start, line.end)  # safety
                start = min(line.end, start + min(trim, slack_before))

            # Trim trailing edge, prioritizing removal of silence.
            if trim > 0:
                available_back = max(0.0, end - start - min_duration)
                back_trim = min(trim, available_back)
                end -= back_trim

            # Respect next line gap.
            next_start = lines[idx + 1].start if idx + 1 < len(lines) else None
            if next_start is not None and end > next_start - gap:
                end = next_start - gap

            # Ensure we never exceed original boundaries.
            if start < line.start:
                start = line.start
            if end > line.end:
                end = line.end

            # Enforce minimum duration.
            if end - start < min_duration:
                target_end = start + min_duration
                if next_start is not None and target_end > next_start - gap:
                    target_end = min(next_start - gap, line.end)
                    start = max(line.start, target_end - min_duration)
                end = max(target_end, start + 0.01)

            start = max(start, 0.0)
            if end <= start:
                end = start + min_duration

            start_with_offset = start + offset
            end_with_offset = end + offset
            if start_with_offset < 0:
                shift = -start_with_offset
                start_with_offset += shift
                end_with_offset += shift
            if end_with_offset <= start_with_offset:
                end_with_offset = start_with_offset + min_duration

            adjusted_line = SubtitleLine(
                index=line.index,
                start=start_with_offset,
                end=end_with_offset,
                text=line.text,
            )
            adjusted.append(adjusted_line)
            prev_end_base = end

        return SubtitleDocument(lines=adjusted)

    def remove_trailing_punctuation(
        self,
        punctuation: str = _TRAILING_PUNCTUATION,
        closing_wrappers: str = _CLOSING_WRAPPERS,
    ) -> "SubtitleDocument":
        cleaned: List[SubtitleLine] = []
        wrapper_set = set(closing_wrappers)
        for line in self.lines:
            text = line.text.rstrip()
            original = text
            if not text:
                cleaned.append(line)
                continue

            trailing_wrappers: List[str] = []
            while text and text[-1] in wrapper_set:
                trailing_wrappers.append(text[-1])
                text = text[:-1].rstrip()

            stripped = text.rstrip(punctuation).rstrip()
            if not stripped:
                stripped = original
                trailing_wrappers.clear()

            rebuilt = stripped + "".join(reversed(trailing_wrappers))
            cleaned.append(
                SubtitleLine(
                    index=line.index,
                    start=line.start,
                    end=line.end,
                    text=rebuilt,
                )
            )
        return SubtitleDocument(lines=cleaned)

    def normalize_cjk_spacing(self) -> "SubtitleDocument":
        normalized: List[SubtitleLine] = []
        for line in self.lines:
            text = line.text
            text = _CJK_LATIN_LEFT_PATTERN.sub(r"\1 \2", text)
            text = _CJK_LATIN_RIGHT_PATTERN.sub(r"\1 \2", text)
            text = re.sub(r" {2,}", " ", text).strip()
            normalized.append(
                SubtitleLine(
                    index=line.index,
                    start=line.start,
                    end=line.end,
                    text=text,
                )
            )
        return SubtitleDocument(lines=normalized)


def format_timestamp(seconds: float, separator: str = ",") -> str:
    """Format seconds into SRT/VTT timestamp format."""

    if seconds < 0:
        seconds = 0
    delta = timedelta(seconds=seconds)
    total_seconds = int(delta.total_seconds())
    hours, remainder = divmod(total_seconds, 3600)
    minutes, secs = divmod(remainder, 60)
    milliseconds = int(delta.microseconds / 1000)
    return f"{hours:02}:{minutes:02}:{secs:02}{separator}{milliseconds:03}"


def parse_timestamp(value: str) -> float:
    """Parse SRT/VTT timestamp into seconds."""

    value = value.strip()
    if not value:
        raise ValueError("Empty timestamp")
    separator = "," if "," in value else "."
    try:
        hours_str, minutes_str, rest = value.split(":", 2)
        seconds_str, millis_str = rest.split(separator)
        hours = int(hours_str)
        minutes = int(minutes_str)
        seconds = int(seconds_str)
        milliseconds = int(millis_str)
    except ValueError as exc:
        raise ValueError(f"Invalid timestamp format: {value}") from exc
    total = hours * 3600 + minutes * 60 + seconds + milliseconds / 1000.0
    return total


__all__ = [
    "SubtitleLine",
    "SubtitleDocument",
    "format_timestamp",
    "parse_timestamp",
]



from __future__ import annotations

import shutil
import subprocess
from pathlib import Path

from .config import PipelineConfig


class AudioExtractionError(RuntimeError):
    """Raised when ffmpeg fails to extract audio."""


FFMPEG_NOT_FOUND = (
    "ffmpeg is required but was not found in PATH. "
    "Install ffmpeg and try again."
)


def ensure_ffmpeg() -> str:
    """Return ffmpeg executable path or raise if missing."""

    ffmpeg_path = shutil.which("ffmpeg")
    if not ffmpeg_path:
        raise FileNotFoundError(FFMPEG_NOT_FOUND)
    return ffmpeg_path


def extract_audio(
    video_path: Path,
    pipeline: PipelineConfig,
    work_dir: Path,
) -> Path:
    """Extract audio from the given video file using ffmpeg."""

    ffmpeg_path = ensure_ffmpeg()
    audio_format = pipeline.audio_format
    work_dir.mkdir(parents=True, exist_ok=True)
    audio_path = work_dir / f"{video_path.stem}.{audio_format}"

    cmd = _build_ffmpeg_cmd(
        ffmpeg_path=ffmpeg_path,
        video_path=video_path,
        audio_path=audio_path,
        audio_format=audio_format,
    )

    result = subprocess.run(
        cmd,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        raise AudioExtractionError(
            f"ffmpeg failed with exit code {result.returncode}: {result.stderr.strip()}"
        )
    return audio_path


def _build_ffmpeg_cmd(
    ffmpeg_path: str, video_path: Path, audio_path: Path, audio_format: str
) -> list[str]:
    """Construct ffmpeg command depending on desired output format."""

    base_cmd = [ffmpeg_path, "-y", "-i", str(video_path), "-vn"]

    if audio_format == "wav":
        base_cmd += ["-ac", "1", "-ar", "16000", "-f", "wav", str(audio_path)]
    elif audio_format == "mp3":
        base_cmd += ["-acodec", "libmp3lame", str(audio_path)]
    elif audio_format == "flac":
        base_cmd += ["-acodec", "flac", str(audio_path)]
    elif audio_format == "m4a":
        base_cmd += ["-acodec", "aac", str(audio_path)]
    elif audio_format == "ogg":
        base_cmd += ["-acodec", "libvorbis", str(audio_path)]
    else:
        raise ValueError(f"Unsupported audio format: {audio_format}")

    return base_cmd


__all__ = [
    "AudioExtractionError",
    "extract_audio",
    "ensure_ffmpeg",
]




from __future__ import annotations

import json
from pathlib import Path
from typing import Dict, Iterable, Optional


class PipelineState:
    """Simple JSON-backed state to support resuming a pipeline run."""

    def __init__(self, path: Path, data: dict) -> None:
        self.path = path
        self._data = data

    @classmethod
    def load(cls, path: Path, video_path: Path) -> "PipelineState":
        if path.exists():
            with path.open("r", encoding="utf-8") as fh:
                data = json.load(fh)
        else:
            data = {
                "video": str(video_path.resolve()),
                "audio": {},
                "transcription": {},
                "translation": {},
            }
        return cls(path, data)

    # ---- Audio -----------------------------------------------------

    def get_audio_path(self) -> Optional[Path]:
        audio = self._data.get("audio", {})
        path = audio.get("path")
        return Path(path) if path else None

    def set_audio_path(self, path: Path) -> None:
        self._data.setdefault("audio", {})["path"] = str(path)
        self.save()

    # ---- Transcription ---------------------------------------------

    def get_segments_path(self) -> Optional[Path]:
        transcription = self._data.get("transcription", {})
        path = transcription.get("segments_path")
        return Path(path) if path else None

    def mark_transcription(
        self, segments_path: Path, total_lines: int
    ) -> None:
        self._data.setdefault("transcription", {}).update(
            {
                "segments_path": str(segments_path),
                "total_lines": total_lines,
            }
        )
        self.save()

    def transcription_total_lines(self) -> Optional[int]:
        transcription = self._data.get("transcription", {})
        return transcription.get("total_lines")

    # ---- Translation -----------------------------------------------

    def translation_progress_path(self, default: Path) -> Path:
        translation = self._data.setdefault("translation", {})
        path = translation.get("progress_path")
        if not path:
            translation["progress_path"] = str(default)
            self.save()
            return default
        return Path(path)

    def completed_lines(self) -> set[str]:
        translation = self._data.get("translation", {})
        return set(translation.get("completed_lines", []))

    def mark_lines_completed(self, line_ids: Iterable[str]) -> None:
        translation = self._data.setdefault("translation", {})
        completed = set(translation.get("completed_lines", []))
        completed.update(line_ids)
        translation["completed_lines"] = sorted(completed, key=int)
        self.save()

    def reset_translation(self) -> None:
        if "translation" in self._data:
            del self._data["translation"]
        self.save()

    # ---- Cleanup ---------------------------------------------------

    def clear(self) -> None:
        if self.path.exists():
            self.path.unlink()

    # ---- Persistence -----------------------------------------------

    def save(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)
        with self.path.open("w", encoding="utf-8") as fh:
            json.dump(self._data, fh, indent=2)


def load_translation_progress(path: Path) -> Dict[str, str]:
    if not path.exists():
        return {}
    with path.open("r", encoding="utf-8") as fh:
        data = json.load(fh)
    return data.get("line_translations", {})


def persist_translation_progress(path: Path, translations: Dict[str, str]) -> None:
    path.parent.mkdir(parents=True, exist_ok=True)
    with path.open("w", encoding="utf-8") as fh:
        json.dump({"line_translations": translations}, fh, ensure_ascii=False, indent=2)


__all__ = ["PipelineState", "load_translation_progress", "persist_translation_progress"]




from __future__ import annotations

import logging
from logging.handlers import RotatingFileHandler
from pathlib import Path

_LOGGER_NAME = "transub"


def setup_logging(log_path: Path) -> logging.Logger:
    """Configure rotating file logging for the CLI run."""

    logger = logging.getLogger(_LOGGER_NAME)
    if logger.handlers:
        return logger

    log_path.parent.mkdir(parents=True, exist_ok=True)
    handler = RotatingFileHandler(
        log_path,
        maxBytes=10 * 1024 * 1024,  # 10MB
        backupCount=3,
        encoding="utf-8",
    )
    formatter = logging.Formatter(
        "%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )
    handler.setFormatter(formatter)

    logger.setLevel(logging.INFO)
    logger.addHandler(handler)
    logger.propagate = False

    return logger


def get_logger() -> logging.Logger:
    """Return the configured Transub logger."""

    return logging.getLogger(_LOGGER_NAME)


__all__ = ["setup_logging", "get_logger"]

